from __future__ import (division, generators, absolute_import,
                        print_function, with_statement, nested_scopes,
                        unicode_literals)
import matplotlib.pyplot as plt
import numpy as np
import pathlib
import pandas as pd
import random
import seaborn as sns
import sys
import warnings

from numba import jit
from numpy import linalg as la
from scipy.special import loggamma
from scipy.stats import chi2
from scipy.linalg import toeplitz, solve
from sklearn.preprocessing import scale
from timeit import default_timer as dt

from .tmath import cov2cor
from .utils import pretty_print_time
from .distributions import qnorm_aprox


def _updateProgBarMCMC(curIter, totalIter, t0, ar, barLength=20):
    '''
    Custom progress bar to output MCMC chain progress.

    Parameters
    ----------
    curIter : int
        Current iteration.
    totalIter : int
        Total iterations.
    t0 : float
        Timestamp of when the process started (timestamp as float).
    ar : float
        Acceptance Ratio.
    barLength : int, optional
        The character length of the progress bar. The default is 20.

    Returns
    -------
    None.

    '''
    status = "Working..."
    progress = float(curIter)/float(totalIter)
    if isinstance(progress, int):
        progress = float(progress)
    if progress >= 1:
        progress = 1
        status = "Finished!..."
    block = int(round(barLength*progress))
    text = "\rPercent: [{0}] {1:.2f}% iter: {2}/{3}, {4} Elapsed: {5}, Est: {6}, Accept. Rate: {7}".format(
        "#"*block + "-"*(barLength - block), 
        round(progress*100.0, 2), curIter, totalIter, status, pretty_print_time(t0, dt()),
        pretty_print_time((dt()-t0)/curIter * (totalIter - curIter)), np.round(ar, 3))
    if progress >= 1:
        sys.stdout.write(text + "\r\n")
        sys.stdout.flush()
    else:
        sys.stdout.write(text)
        sys.stdout.flush()
        

def applyMCMC(st, ni, lp, algo, algoOpts=None, postArgs={},
              sd=0.02, max_tries=100):
    '''
    This function iteratively applies the MCMC initialization. Since the MCMC
    algorithms used here involve a cholesky decomposition, the methods
    sometimes get stuck with a covaraince matrix that is not positive definite.
    This will attempt to jitter the covariance matrix until it can initialize
    properly.

    Parameters
    ----------
    st : numpy array
        An array of the parameter starting values.
    ni : int
        NUmber of MCMC iterations.
    lp : function
        Function for the log posterior.
    algo : function
        MCMC algorithm to be performed.
    algoOpts : dict, optional
        Specific options for the MCMC algorithm. The default is None.
    postArgs : dict, optional
        Specific options for the posterior function. The default is None.
    sd : float, optional
        The standard deviation of the normal distribution used to draw the
        jitter amount from. In other words, the jittered covariance is the 
        covaraince matrix plus a random draw X, where X~N(0, sd). 
        The default is 0.02.
    max_tries : int, optional
        The max number of times to try and jitter before admitting defeat.
        If the jitter fails, the reason or the covaraince matrix not being
        positive definite may not be due to randomness, and may require
        a re-evaluation of the problem space. The default is 100.

    Raises
    ------
    ValueError
        Raised when cholesky decomposition fails after max_tries.

    Returns
    -------
    res : tuple
        Returns tuple containing the MCMC results.

    '''
    try_num = 1
    not_successful = True
    res = None
    lns = st.shape
    while not_successful:
        if try_num % 5 == 0:
            st = st + np.random.normal(size=lns, scale=sd)
        try:
            res = algo(start=st, niter=ni, lpost=lp, postArgs=postArgs,
                       options=algoOpts)
            not_successful = False
        except np.linalg.LinAlgError:
            try_num += 1
        
        if try_num >= max_tries:
            raise ValueError("Cholesky Decomposition was not successful after " + str(max_tries) + " tries. Try new starting values")
    print("Number of Cholesky tries: " + str(try_num))
    return res              


# For upper triangle rank one update
@jit
def cholupdate(L, x, update=True):
    '''
    Upper triangle, rank one update for cholesky decomposed matrix. 

    Parameters
    ----------
    L : numpy array (float)
        The upper-triangular decomposed matrix, shape=(N, N).
    x : numpy array (float)
        The values being added to L, shape=(N, ).
    update : bool, optional
        Perform an update (as opposed to a downdate). The default is True.

    Returns
    -------
    L : numpy array
        Return updated L matrix.

    '''
    p = len(x)
    for k in range(p):
        if update:
            r = np.sqrt((L[k, k]**2) + (x[k]**2))
        else:
            r = np.sqrt((L[k, k]**2) - (x[k]**2))
        c = r / L[k, k]
        s = x[k] / L[k, k]
        L[k, k] = r
        if k < (p - 1):
            if update:
                L[k, (k + 1):p] = (L[k, (k + 1):p] + s * x[(k + 1):p]) / c
            else:
                L[k, (k + 1):p] = (L[k, (k + 1):p] - s * x[(k + 1):p]) / c
            x[(k + 1):p] = c * x[(k + 1):p] - s * L[k, (k + 1):p]
    return L


def adaptive_mcmc(start, niter, lpost, postArgs={}, options=None):
    '''
    A random walk metropolis algorithm that adaptively tunes the covaraince 
    matrix. Based on methods by Rosenthal (who improved on Haario's method).
    The method by Rosenthal is sometimes refered to as Adaptive Mixture
    Metropolis, while the algorithm by Haario is called Adaptive Metropolis and
    is generally considered to be the historically first adaptive Metropolis
    algorithm.

    Parameters
    ----------
    start : numpy array
        Starting values for the MCMC.
    niter : int
        Number of iterations.
    lpost : function
        Log posterior function.
    postArgs : dict
        Extra arguments for the log posterior function. If there are none, pass
        an empty dictionary.
    options : dict, optional
        Extra arguments for the MCMC algorithm, namely:
            beta : float
                Between 0 and 1. Decides the proportion to sample for each
                section of the mixture distribution.
                
                A mixture distribution is essentially like adding two 
                distributions together. However, to avoid some complicated
                math, one way to sample from a mixture of two distributions is
                to use a trick, namely, to first sample from a uniform
                distribution between 0, 1, and then evaluate whether that value
                is above some threshold (beta in this case). If it is, sample 
                from the first distribution, otherwise, sample from the second.
            progress : bool
                Whether to display progress bar
            prev_vals : dict
                The previous values of the last run, namely:
                    chol2 : numpy array
                        the decomposed covariance matrix of the parameters
                    sumx : numpy array
                        the current sum of the parameter value (for each
                                                                parameter) 
                    prev_i : int or float
                        the number of samples represented in sumx.
                        Used in averaging sumx
        The default is None.

    Returns
    -------
    parm : numpy array
        MCMC samples.
    prev_vals : dict
        The ending values of the MCMC algorithm. Useful when you want to
        continue where you left off.

    '''
    
    beta = 0.05
    progress = True
    prev_vals = {'chol2': None, 'sumx': 0.0, 'prev_i': 0.0}
    if options is not None:
        keys = list(options.keys())
        if 'beta' in keys:
            beta = options['beta']
        if 'progress' in keys:
            progress = options['progress']
        if 'prev_vals' in keys:
            prev_vals.update(options['prev_vals'])
    
    numParams = start.size
    sqrtNumParams = np.sqrt(numParams)
    parm = np.zeros(shape=(niter, numParams))
    parm[0, ] = start
    sumx = start + prev_vals['sumx']
    accept = 0
    post_old = lpost(start, **postArgs)
    
    prop_dist_var = (0.1**2) * np.diag(np.repeat(1, numParams)) / numParams
    chol1 = la.cholesky(prop_dist_var)
    chol2 = prev_vals['chol2']
    acceptDraw = False
    loop = range(1, niter)
    
    sumi = 1.0 + prev_vals['prev_i']
    t0 = dt()
    for i in loop:
        parm[i, ] = parm[i - 1, ]
        
        if i <= ((2 * numParams) - 1):
            tune = chol1
        else:
            if chol2 is None:
                XXt = parm[0:i, ].T.dot(parm[0:i, ])
                chol2 = la.cholesky(XXt).T
            else:
                chol2 = cholupdate(chol2, np.array(parm[i - 1, ]))
            
            if random.random() < beta:
                tune = chol1
            else:
                tune = (2.38*cholupdate(chol2 / np.sqrt(sumi), sumx/sumi, update=False) / sqrtNumParams * np.sqrt(sumi / (sumi - 1)))
        
        if np.any(np.isnan(tune)):
            tune = chol1
        cand = np.random.normal(size=numParams).dot(tune) + parm[i - 1, ]
        post_new = lpost(cand, **postArgs)
        
        if (post_new - post_old) > np.log(random.random()):
            acceptDraw = True
        
        if acceptDraw:
            parm[i, ] = cand
            post_old = post_new
            accept += 1
        
        sumx = sumx + parm[i, ]
        sumi += 1.0
        acceptDraw = False
        if progress:
            _updateProgBarMCMC(i + 1, niter, t0, float(accept) / float(i))
        
    prev_vals = {'chol2': chol2, 'prev_i': sumi - 1, 'sumx': sumx}
    print("Acceptance Rate: ", float(accept) / float(niter))
    return {'parameters': parm, 'prev_vals': prev_vals}


def rwm_with_lap(start, niter, lpost, postArgs={}, options=None):
    '''
    A random walk metropolis algorithm that adaptively tunes the covaraince 
    matrix with a log-adaptive posterior.
    
    See "Exploring an Adaptive Metropolis Algorithm" by Ben Shaby, 2010.

    Parameters
    ----------
    start : numpy array
        Starting values for the MCMC.
    niter : int
        Number of iterations.
    lpost : function
        Log posterior function.
    postArgs : dict
        Extra arguments for the log posterior function. The default is
        an empty dictionary.
    options : dict, optional
        Extra arguments for the MCMC algorithm, namely:
            k : int
                The number of MCMC samples to generate for each evaluation.
            c0 : float
                Attenuation parameter. Default is 1.
            c1 : float
                Attenuation parameter. Default is 0.8.
            progress : bool
                Whether to display progress bar
            prev_vals : dict
                The previous values of the last run, namely:
                    E_0 : numpy array
                        the final covaraince matrix
                    sigma_2 : float
                        the positive scaling parameter in the algorithm
                    t : int
                        the current iteration number
        The default is None.

    Returns
    -------
    parm : numpy array
        MCMC samples.
    prev_vals : dict
        The ending values of the MCMC algorithm. Useful when you want to
        continue where you left off.

    '''
    k = 20
    c_0 = 1.0
    c_1 = 0.8
    progress = True
    prev_vals = {'E_0': None, 'sigma_2': None, 't': 0.0}
    if options is not None:
        keys = list(options.keys())
        if 'k' in keys:
            k = options['k']
        if 'c_0' in keys:
            c_0 = options['c_0']
        if 'c_1' in keys:
            c_1 = options['c_1']
        if 'progress' in keys:
            progress = options['progress']
        if 'prev_vals' in keys:
            prev_vals.update(options['prev_vals'])
    
    numParams = start.size
    optimal = 0.444
    if numParams >= 2:
        optimal = 0.234
    T_iter = np.ceil(niter/float(k))
    niter = int(T_iter * k)
    parm = np.zeros(shape=(niter, numParams))
    parm[0, ] = start
    
    total_accept = k_accept = 0
    post_old = lpost(start, **postArgs)
    
    sigma_2 = (2.38**2)/numParams
    if prev_vals['sigma_2'] is not None:
        sigma_2 = prev_vals["sigma_2"]
    
    E_0 = np.diag(np.repeat(1, numParams))
    if prev_vals['E_0'] is not None:
        E_0 = prev_vals["E_0"]

    chol = la.cholesky(np.sqrt(sigma_2)*E_0)
    chol_i = np.array(chol)
    
    t = 1 + prev_vals['t']
    
    acceptDraw = False
    loop = range(1, niter)
    
    t0 = dt()
    for i in loop:
        parm[i, ] = parm[i - 1, ]
        cand = np.random.normal(size=numParams).dot(chol) + parm[i - 1, ]
        
        post_new = lpost(cand, **postArgs)
        
        if (post_new - post_old) > np.log(random.random()):
            acceptDraw = True
        
        if acceptDraw:
            parm[i, ] = cand
            post_old = post_new
            k_accept += 1
            total_accept += 1
            
        acceptDraw = False
        if progress:
            _updateProgBarMCMC(i + 1, niter, t0, float(total_accept) / float(i))
        
        if (i + 1) % k == 0:
            X = parm[(i + 1 - k):(i + 1), :]
            mean_X = np.mean(X, axis=0)
            
            r_t = k_accept / float(k)
            Ehat_0 = (1.0 / (k - 1.0)) * ((X - mean_X).T.dot((X - mean_X)))
            gamma_1 = 1/(t**c_1)
            gamma_2 = c_0 * gamma_1
            sigma_2 = np.exp(np.log(sigma_2) + (gamma_2 * (r_t - optimal)))
            E_0 = E_0 + gamma_1*(Ehat_0 - E_0)
            
            if np.any(np.isnan(E_0)) or not np.all(np.isfinite(E_0)):
                chol = chol_i
            else:
                try:
                    chol = la.cholesky(np.sqrt(sigma_2)*E_0)
                #except la.LinAlgError:
                #    chol = sla.sqrtm(sigma_2*E_0)
                except:
                    chol = chol_i
                    
            t += 1
            k_accept = 0
            
    prev_vals = {'E_0': E_0, 'sigma_2': sigma_2, 't': t}
    print("Acceptance Rate: ", float(total_accept) / float(niter))
    return {'parameters': parm, 'prev_vals': prev_vals}


def rwm(start, niter, lpost, postArgs={}, options=None):
    '''
    A random walk metropolis algorithm.

    Parameters
    ----------
    start : numpy array
        Starting values for the MCMC.
    niter : int
        Number of iterations.
    lpost : function
        Log posterior function.
    postArgs : dict
        Extra arguments for the log posterior function. The default is
        an empty dictionary.
    options : dict, optional
        Extra arguments for the MCMC algorithm, namely:
            E : numpy array
                The covariance matrix
            progress : bool
                Whether to display progress bar
            prev_vals : dict
                The previous values of the last run, namely:
                    E_0 : numpy array
                        the final covaraince matrix
        The default is None.

    Returns
    -------
    parm : numpy array
        MCMC samples.
    prev_vals : dict
        The ending values of the MCMC algorithm. Useful when you want to
        continue where you left off.

    '''
    numParams = start.size
    
    prev_vals = {'E_0': ((2.38**2)/numParams)*np.diag(np.repeat(1, numParams))}
    progress = True
    if options is not None:
        keys = list(options.keys())
        if 'E' in keys:
            E = options['E']
        if 'progress' in keys:
            progress = options['progress']
        if 'prev_vals' in keys:
            prev_vals.update(options['prev_vals'])
    
    parm = np.zeros(shape=(niter, numParams))
    parm[0, ] = start
    
    accept = 0
    post_old = lpost(start, **postArgs)
    
    E = prev_vals['E_0']
    chol = la.cholesky(E)
    
    acceptDraw = False
    loop = range(1, niter)
    
    t0 = dt()
    for i in loop:
        parm[i, ] = parm[i - 1, ]
        cand = np.random.normal(size=numParams).dot(chol) + parm[i - 1, ]
        
        post_new = lpost(cand, **postArgs)
        
        if (post_new - post_old) > np.log(random.random()):
            acceptDraw = True
        
        if acceptDraw:
            parm[i, ] = cand
            post_old = post_new
            accept += 1
        
        acceptDraw = False
        if progress:
            _updateProgBarMCMC(i + 1, niter, t0, float(accept) / float(i))
    
    prev_vals = {'E_O': E}
    print("Acceptance Rate: ", float(accept) / float(niter))
    return {'parameters': parm, 'prev_vals': prev_vals}


def samp_size_calc_raftery(q=0.025, r=0.005, s=0.95):
    '''
    Calculate needed sample size for Raftery evaluation.

    Parameters
    ----------
    q : float, optional
        Quantile of interest (in terms of percentile, i.e. between 0 and 1).
        The default is 0.025.
    r : float, optional
        Accuracy. The default is 0.005.
    s : float, optional
        Probability. The default is 0.95.

    Returns
    -------
    phi : float
        Phi parameter in Raftery evaluation.
    nmin : int
        Minimum number of samples needed.

    '''
    phi = qnorm_aprox(0.5 * (1.0 + s))
    nmin = np.ceil((q * (1.0 - q) * phi**2)/r**2).astype(int)
    return phi, nmin


def lag(s, b, method):
    '''
    Translated from R's mcmcse package
    
    Returns the lag window value for the corresponding window.

    Parameters
    ----------
    TODO: add description for these parameters
    s : int
        DESCRIPTION.
    b : flost
        DESCRIPTION.
    method : str
        Either `bartlett` or None.

    Returns
    -------
    int, float
        Lag window.

    '''
    if method == "bartlett":
        return 1 - (s/b)
    else:
        return ((1 + np.cos(np.pi * s/b))/2) 


def adjust_matrix(mat, N, epsilon=None, b=9/10):
    '''
    Translated from R's mcmcse package.
    
    Function adjusts a non-positive definite estimator to be positive definite.

    Parameters
    ----------
    mat : numpy array
        A symmetric pxp matrix, usually a covarince matrix.
    N : int
        Number of observations in the original atrix.
    epsilon : float, optional
        The adjustment size. If None, sqrt(log(N)/p).
        The default is None.
    b : float, optional
        The exponent on N for the adjustment. The default is 9/10.

    Returns
    -------
    mat_adj : numpy array
        Adjusted matrix.

    '''
    if epsilon is None:
        epsilon = np.sqrt(np.log(N)/mat.shape[1])
    mat_adj = mat
    adj = epsilon*N**(-b)
    var = np.diag(mat)
    corr = cov2cor(mat)
    eig_val, eig_vec = np.linalg.eig(corr)
    adj_eigs = np.maximum(eig_val, adj)
    mat_adj = np.diag(var**0.5).dot(eig_vec).dot(np.diag(adj_eigs)).dot(eig_vec.T).dot(np.diag(var**0.5))
    return mat_adj


#def arp_approx(chain):
#
#
#
#def batchSize(chain, method="bm", g=None):
#    if g is not None:
#        chain = np.array([g(chain[i, :]) for i in range(chain.shape[0])])
#    n = chain.shape[0]
    


def mbmc(chain, b):
    '''
    Translated from R's mcmcse package

    Parameters
    ----------
    chain : numpy array
        MCMC chain.
    b : int
        Number of blocks.

    Returns
    -------
    numpy array
        Covaraince matrix estimate.

    '''
    b = int(b)
    n = chain.shape[0]
    chain = chain.reshape(n, -1)
    p = chain.shape[1]
    a = int(np.floor(n/b))
    y_mean = np.zeros(p)
    out = np.zeros((p, p))
    block_means = np.zeros((a, p))
    mean_mat = np.zeros((a, p))
    idx = np.arange(a) * b

    for i in range(b):
        block_means += chain[idx, :]
        idx += 1
    block_means = block_means/b

    y_mean = np.mean(chain, axis=0)
    for i in range(a):
        mean_mat[i, :] = y_mean

    out = (block_means - mean_mat).T.dot(block_means - mean_mat)
    return (out*b/(a - 1))


def mobmc(chain, b):
    '''
    Translated from R's mcmcse package

    Parameters
    ----------
    chain : numpy array
        MCMC chain.
    b : int
        Number of blocks.

    Returns
    -------
    numpy array
        Covariance matrix estimate.

    '''
    b = int(b)
    n = chain.shape[0]
    chain = chain.reshape(n, -1)
    p = chain.shape[1]
    a = n - b + 1
    
    y_mean = np.zeros(p)
    out = np.zeros((p, p))
    block_means = np.zeros((a, p))
    mean_mat = np.zeros((a, p))
    
    idx = np.arange(a)

    for i in range(b):
        block_means += chain[idx, :]
        idx += 1
    block_means = block_means/b

    y_mean = np.mean(chain, axis=0)
    for i in range(a):
        mean_mat[i, :] = y_mean

    out = (block_means - mean_mat).T.dot(block_means - mean_mat)
    return (out*b/n)


def msvec(chain, b, method="bartlett"):
    '''
    Translated from R's mcmcse package.

    Parameters
    ----------
    chain : numpy array
        MCMC chain.
    b : int
        Number of blocks.
    method : str, optional
        Method to estimate covariance matrix. The default is "bartlett".

    Returns
    -------
    numpy array
        Covariance matrix estimate.

    '''
    n = chain.shape[0]
    chain = chain.reshape(n, -1)
    p = chain.shape[1]
    tchain = chain.T
    out = np.zeros((p, p))
    dummy = np.zeros((p, p))

    for s in range(int(b)):
        dummy = tchain[:, 0:(n-s-1)].dot(chain[s:n-1, :])
        out += lag(s, b, method)*(dummy + dummy.T)

    out += tchain.dot(chain)
    return out/n


def mcse_multi(chain, method="bm", r=3, size="sqroot", g=None, adjust=True):
    '''
    Translated from R's mcmcse package.
    
    An estimate of the Monte Carlo Standard Error, as well as the Monte Carlo
    estimate. Returns a covariance matrix and array for the estimates, as well 
    as other algorithmic outputs.

    Parameters
    ----------
    chain : numpy array
        The MCMC chain, where the rows are samples.
    method : str, optional
        Any of `bm`, `obm`, `bartlett`, `tukey`. `bm` represents batch means
        estimator, `obm` represents overlapping batch means estimator with,
        `bartlett` and `tukey` represents the modified-Bartlett window and
        the Tukey-Hanning windows for spectral variance estimators.
        The default is "bm".
    r : int, float, optional
        The Lugsail parameter, which converts a lag window into it's lugsail
        equivalent. Larger r implies less underestimation of `cov`, but higher
        variability of the estimator. r > 5 is not recommended.
        The default is 3.
    size : str, or int, optional
        Batch size, either `sqroot`, `cuberoot`, or an int value between 1 and
        n/2. 
        TODO: switch default to None once batch_size is implemented.
        The default is 'sqroot'.
    g : function, optional
        A function to apply to the samples of the chain. If None,
        g is set to be the identity, which is the estimation of the mean of
        the target density. The default is None.
    adjust : bool, optional
        Automatically adjust the matrix if it is unstable.
        The default is True.

    Raises
    ------
    ValueError
        Raised is if r is negative, if size is misspecified, if b and r both
        equal 1, or if an unknown method is specified.

    Returns
    -------
    dict
        A dictionary of the results.

    '''
    method_used = method
    if method == "lug":
        method = "bm"
        r = 3
    if r > 5:
        warnings.warn("""It's recommended to use r <= 5. Also, r cannot be
                      negative""")
    if r < 0:
        raise ValueError("r cannot be negative.")
    if g is not None:
        chain = np.array([g(chain[i, :]) for i in range(chain.shape[0])])
    n = chain.shape[0]
    chain = chain.reshape(n, -1)
    p = chain.shape[1]
    # TODO: include batchSize function as option in the ifelse below (see 
    # mcmcse R docs)
    #if size is None:
    #    batchSize(chain=chain, method=method, g=g)
    if size == "sqroot":
        b = np.floor(np.sqrt(n))
    elif size == "cuberoot":
        b = np.floor(n**(1/3))
    else:
        if size < 1 or size >= n or np.floor(n/size) <= 1:
            raise ValueError("""'size' must be a numeric quantity not
                             larger than n.""")
        b = np.floor(size)

    if b == 1 and r != 1:
        r = 1
        message = "r was set to 1 since b = 1."
    mu_hat = np.mean(chain, axis=0)
    sig_mat = np.zeros(0, (p, p))
    if np.floor(b/r) < 1:
        raise ValueError("Either decrease r or increase n")
    message = ""
    
    if method != "bm" and method != "obm" and method != "bartlett" and method != "tukey":
        raise ValueError("No such method available")
    elif method == "bm":
        bm_mat = mbmc(chain, b)
        sig_mat = bm_mat
        method_used = "Batch Means"
        if r > 1:
            sig_mat = 2*bm_mat - mbmc(chain, np.floor(b/r))
            method_used <- "Lugsail Batch Means with r = " + str(r)
            if np.prod(np.diag(sig_mat) > 0) == 0:
                sig_mat = bm_mat
                method_used = "Batch Means"
                message = "Diagonals were negative with r = " + str(r) + ". r = 1 was used."
    elif method == "obm":         
        obm_mat = mobmc(chain, b)
        sig_mat = obm_mat
        method_used = "Overlapping Batch Means"
        if r > 1:
            sig_mat = 2*obm_mat - mobmc(chain, np.floor(b/r))
            method_used <- "Lugsail Overlapping Batch Means with r = " + str(r)
            if np.prod(np.diag(sig_mat) > 0) == 0:
                sig_mat = obm_mat
                method_used = "Overlapping Batch Means"
                message = "Diagonals were negative with r = " + str(r) + ". r = 1 was used."
    elif method == "bartlett":
        chain = scale(chain, with_mean=True, with_std=False)
        bar_mat = msvec(chain, b, "bartlett")
        sig_mat = bar_mat
        method_used = "Bartlett Spectral Variance"
        if r > 1:
            sig_mat = 2*bar_mat - msvec(chain, np.floor(b/r), "bartlett")
            method_used = "Lugsail Bartlett Spectral Variance with r = " + str(r)
            if np.prod(np.diag(sig_mat) > 0) == 0:
                sig_mat = bar_mat
                method_used = "Bartlett Spectral Variance"
                message = "Diagonals were negative with r = " + str(r) + ". r = 1 was used."
    
    elif method == "tukey":
        chain = scale(chain, with_mean=True, with_std=False)
        tuk_mat = msvec(chain, b, "tukey")
        method_used = "Tukey Spectral Variance"
        if r > 1:
            sig_mat = 2*tuk_mat - msvec(chain, np.floor(b/r), "tukey")
            method_used = "Lugsail Tukey Spectral Variance with r = " + str(r)
            if np.prod(np.diag(sig_mat) > 0) == 0:
                sig_mat = tuk_mat
                method_used = "Tukey Spectral Variance"
                message = "Diagonals were negative with r = " + str(r) + ". r = 1 was used."
    
    adjust_used = False
    if adjust:
        sig_eigen = np.linalg.eigvals(sig_mat)
        if (min(sig_eigen) <= 0):
            adjust_used = True
            sig_mat = adjust_matrix(sig_mat, N=n)

    return {'cov': sig_mat, 'est': mu_hat, 'nsim': n,
            'size': b, 'adjustment_used': adjust_used,
            'method': method, 'method_used': method_used,
            'message': message}


def minESS(p, alpha=0.05, eps=0.05, ess=None):
    '''
    Translated from the R mcmcse package.
    
    Calculates the minimum Effective Sample Size, independent of the MCMC
    chain for the given number of parameters. `alpha` is the confidence level,
    `eps` is the tolerance level (ignored when `ess is not None`), and `ess`
    is the effective sample size. When `ess is not None`, the function returns
    the tolerance level needed to obtain that ESS.
    
    In practice, the user should find the minESS amount and then sample until 
    they hit that number. Usually, it is computationally difficult to obtain
    the optimal minimum effective sample size, therefore, it is useful to know
    what tolerance is needed to obtain the samples that can be afforded 
    computationally.

    see mcmcse::minESS for more information.

    Parameters
    ----------
    p : int
        The dimension of the estimation problem (i.e. the number of parameters
        represented in the MCMC chain, or the number of columns in the MCMC
        chain).
    alpha : float, optional
        Confidence level. The default is 0.05.
    eps : float, optional
        Tolerance level. The smaller the tolerance, the larger the minimum 
        effective samples. The default is 0.05.
    ess : int, optional
        Estimated effective sample size. The default is None.

    Returns
    -------
    int
        The minimum effective sample required for a given eps tolerance.
        If ess is specified, then the value returned is the eps corresponding
        to that ess.

    '''
    crit = chi2.ppf(1 - alpha, p)
    p2 = 2/p
    if ess is None:
        logminESS = p2 * np.log(2) + np.log(np.pi) - p2 * np.log(p) - p2 * loggamma(p/2) - 2 * np.log(eps) + np.log(crit)
        return np.round(np.exp(logminESS))
    else:
        logEPS = 0.5 * p2 * np.log(2) + 0.5 * np.log(np.pi) - 0.5 * p2 * np.log(p) - 0.5 * p2 * loggamma(p/2) - 0.5 * np.log(ess) + 0.5 * np.log(crit)
        return np.exp(logEPS)


def multiESS(chain, covmat=None, g=None, mcse_multi_args={}):
    '''
    This function computes the Effective Sample Size of an MCMC chain. Due to
    correlation between MCMC samples, it is sometimes unclear how much
    information about the parameters has been obtained. If all of the MCMC
    samples were independent, we would need less samples to get accurate
    information about the posterior than when the samples are correlated.
    ESS measures the amount of independent samples that have actually been
    obtained in the MCMC chain, and mESS is a special case for multivariate
    posteriors. In other words, this method is a way to test if your chain
    has gone far enough.
    
    This information can used in conjunction with minESS, such that the chain
    has sampled enough when multiESS >= minESS.
    
    For more information regarding these functions, see the R documentation.

    Parameters
    ----------
    chain : numpy array
        The MCMC chain, where the rows are samples.
    covmat : numpy array, optional
        The covaraince matrix for the parameters, if available. If None,
        matrix is obtained from mcse_multi. The default is None.
    g : function, optional
        A function that represents features of
        interest. `g` is applied to each row of x, and should thus take a
        vector input only. If g is none, g is set to be identity, which is
        estimation of the mean of the target density. The default is None.
        
        An example of g would be the sum of the second moments of
        each parameter, i.e.:
        
        def g(x):
            return np.sum(x**2)
        
    mcse_multi_args : dict
        Arguments for mcse_multi function. Don't use this if a suitable matrix
        estimate from mcse_multi or mcse_initseq is already obtained. The
        default is an empty dictionary

    Returns
    -------
    ess : int
        The estimated effective sample size.

    '''
    if g is not None:
        chain = np.array([g(chain[i, :]) for i in range(chain.shape[0])])
    
    n = chain.shape[0]
    chain = chain.reshape(n, -1)
    p = chain.shape[1]
    var_mat = np.cov(chain, rowvar=False, ddof=1)
    if covmat is None:
        covmat = mcse_multi(chain, **mcse_multi_args)['cov']
    det_var_p = np.prod(np.linalg.eigvals(var_mat)**(1/p))
    det_covmat_p = np.prod(np.linalg.eigvals(covmat)**(1/p))
    ess = n * (det_var_p/det_covmat_p)
    return ess


def raftery(chain, q=0.025, r=0.005, s=0.95, converge_eps=0.001,
             thin=1, print_=False):
    '''
    Calculate the Raftery diagnostic to determine how many more samples are
    needed.

    Parameters
    ----------
    chain : numpy array
        MCMC chain.
    q : float, optional
        Quantile of interest (in terms of percentile, i.e. between 0 and 1).
        The default is 0.025.
    r : float, optional
        Accuracy. The default is 0.005.
    s : float, optional
        Probability. The default is 0.95.
    converge_eps : float, optional
        Convergence threshold (epsilon). The default is 0.001.
    thin : int, optional
        Thining amount. The default is 1.
    print_ : bool, optional
        Print results. The default is False.

    Raises
    ------
    ValueError
        Raised if there are not enough samples in the chain, given the q,r,s
        values, or if there is an invalid selection of q.

    Returns
    -------
    None.

    '''

    if not q > 0 or not q < 1:
        raise ValueError("q must be between 0 and 1")
    else:
        # forces chain to be shape (:,1) if it is a single parameter chain
        chain = chain.reshape(chain.shape[0], -1)
        niter, nvar = chain.shape
        columns = ["Burn-in (M)", "Total (N)", "Lower Bound (Nmin)",
                   "Dependence Factor (I)", "Thinning (k)"]
        resmatrix = np.empty(shape=(nvar, len(columns)))
        resmatrix[:] = np.nan
        # nmin based on sample size calculation for proportions
        phi = qnorm_aprox(0.5 * (1 + s))
        nmin = np.ceil((q * (1 - q) * phi**2)/r**2).astype(int)
        if (nmin > niter):
            raise ValueError("Error: You either need to adjust q, r, or " +
                             " s, or you need a longer chain (at least " +
                             str(nmin) + " iterations)" +
                             "\r\nInput Parameters: \r\n" +
                             "\tQuantile (q): " + str(q) + "\r\n" +
                             "\tAccuracy (r): +/- " + str(r) + "\r\n" +
                             "\tProbability (s): " + str(s)
                             )
        else:
            for i in range(nvar):
                # U_t = value of the parameter at iteration t
                # u = value of parameter at defined quantile q
                quant = np.percentile(chain[:, i], q=q*100)
                dichot = chain[:, i] <= quant

                kthin = 0
                bic = 1
                testres = None
                newdim = 0
                # To find k ...
                while bic >= 0:
                    kthin = kthin + thin
                    # Z_t, the indicater function, or a dichotomus
                    # variable, representing where U_t <= u
                    testres = dichot[::kthin]
                    newdim = len(testres)

                    testres = pd.Categorical(testres,
                                             categories=[False, True])
                    testtran = pd.crosstab(index=testres[0:(newdim - 2)],
                                           columns=[testres[2:(newdim)],
                                           testres[1:(newdim - 1)]],
                                           margins=False, dropna=False)
                    g2 = 0

                    # First order MC vs. second order MC test (log
                    # likelihood ratio statistic, Bishop, Fienberg and
                    # Holland (1975))
                    for i1 in range(2):
                        for i2 in range(2):
                            for i3 in range(2):
                                if testtran[i1][i2][i3] != 0:
                                    fitted = (float(np.sum(testtran[i1][i2][:]))
                                              * np.sum(testtran[:][i2][i3])
                                              ) / (np.sum(
                                                     np.sum(testtran[:][i2][:])
                                                     ))
                                    g2 = g2 + testtran[i1][i2][i3] * np.log(
                                                 testtran[i1][i2][i3]/fitted
                                                 ) * 2.0
                    bic = g2 - np.log(newdim - 2.0) * 2.0

                finaltran = pd.crosstab(testres[0:(newdim - 1)],
                                        testres[1:(newdim)], dropna=False)
                alpha = finaltran[1][0]/float(finaltran[0][0] + finaltran[1][0])
                beta = finaltran[0][1]/float(finaltran[0][1] + finaltran[1][1])
                tempburn = (np.log((converge_eps * (alpha + beta))
                            / max([alpha, beta]))
                            / (np.log(np.abs(1.0 - alpha - beta)))
                            )

                # M = M * k
                nburn = np.ceil(tempburn) * float(kthin)
                tempprec = (((2.0 - alpha - beta) * alpha *
                             beta * phi**2) /
                            (((alpha + beta)**3) * r**2))
                nkeep = np.ceil(tempprec) * kthin
                # (M+N) / Nmin, which is the increase in the number of
                # iterations due to dependence in the sequence.
                # If I > 1 by a large amount, there is a high level of
                # dependence (rule of thumb: > 5 indicate problems)
                # Problems could be due to bad starting values, high
                # posterior correlations (which are remedied by
                # transformations that remove correlations), or
                # "stickiness" in the chain ( could be resolved by changing
                # MCMC algorithm)
                iratio = (nburn + nkeep)/nmin
                resmatrix[i, 0] = nburn
                resmatrix[i, 1] = nkeep + nburn
                resmatrix[i, 2] = nmin
                resmatrix[i, 3] = np.round(iratio, 2)
                resmatrix[i, 4] = kthin
        df = pd.DataFrame(data=resmatrix, columns=columns)
        inputs = {"r": r, "s": s, "q": q}

        if print_:
            print("\r\nInput Parameters: ", "\r\n",
                  "\tQuantile (q): ", inputs["q"], "\r\n",
                  "\tAccuracy (r): +/-", inputs["r"], "\r\n",
                  "\tProbability (s): ", inputs["s"], "\r\n",
                  "\r\n", df)
        return(df)


class mcmcObject(object):
    '''
    An object to hold MCMC chains, and to store/compute useful metrics on them.
    Also has some common plotting functionality.
    
    To get a chain up and running, run the mcmcWithRaftery method after
    instantiating the mcmcObject class.
    
    '''
    def __init__(self, name="MCMC Object"):
        '''
        Constructor for MCMC object

        Parameters
        ----------
        name : str, optional
            The name of the object (in case you instantiate multiple objects).
            The default is "MCMC Object".

        Returns
        -------
        None.

        '''
        self.name = name
        self.chains = {}
        self.diagnostic_results = {}
        self.previous_values = {}

    def addChain(self, newChain, chainName=None, concat=False):
        '''
        Adds a chain to your collection of chains.

        Parameters
        ----------
        newChain : numpy array
            New chain you would like to add to the collection.
        chainName : str, optional
            The name of the added chain, used to seperate it from others in the
            collection. If None, one will be selected for you.
            The default is None.
        concat : bool, optional
            Whether or not to append the new chain to an existing chain.
            The default is False.

        Returns
        -------
        None.

        '''
        if chainName is None:
            chainName = ''.join(("Chain_", str(len(self.chains) + 1)))

        if not isinstance(newChain, np.ndarray):
            try:
                newChain = np.array(newChain)
            except (ValueError, IndexError, KeyError, TypeError):
                print("Error: Please convert new chain object to Numpy Array")

        if not concat:
            self.chains[chainName] = newChain
        else:
            try:
                self.chains[chainName] = np.concatenate((self.chains[chainName],
                                                         newChain))
            except (NameError, KeyError):
                warnings.warn('''Failed to concatenate chains. Created new
                              chain instead. Check list of chain keys.''')
                self.chains[chainName] = newChain

    def removeChain(self, chainName, print_=True):
        '''
        Remove a chain from the collection.

        Parameters
        ----------
        chainName : str
            Chain to remove.
        print_ : bool, optional
            Print a message displaying what was dropped. The default is True.

        Returns
        -------
        None.

        '''
        try:
            del self.chains[chainName]
            if print_:
                print("Chain called " + chainName + " removed")
        except KeyError:
            print("No chain named ", chainName)

    def showChain(self, chainName):
        '''
        Display a chain from the collection.

        Parameters
        ----------
        chainName : str
            The name of the chain from the collection to display.

        Returns
        -------
        None.

        '''
        try:
            print(self.chains[chainName])
        except KeyError:
            print("No chain named ", chainName)

    def burnin(self, chainName, burninVal=3000, replace=False):
        '''
        Remove values from chain through burnin process (i.e. remove frist
        burninVal samples)

        Parameters
        ----------
        chainName : str
            The name of the chain from the collection.
        burninVal : int, optional
            The number of samples to remove. The default is 3000.
        replace : bool
            If True, replace the current chain rather than create a new one
            without the burnin samples. If a new chain is created, it will
            be called chainName + '_burnin' + burnunVal, and will be available
            in the collection.

        Returns
        -------
        None.

        '''
        if replace:
            self.chains[chainName] = self.chains[chainName][burninVal:, :]
        else:
            self.chains[chainName+'_burnin'+str(burninVal)] = self.chains[chainName][burninVal:, :]

    def bestRaftery(self, chainName, q=[0.025, 0.5, 0.975],
                    r=0.005, s=0.90, converge_eps=0.001, thin=1,
                    print_each=False, print_final=False):
        '''
        Run multiple Raftery evaluations and compare. The final raftery output
        is the maximum value for that criteria obtained from all Raftery
        evaluations.

        Parameters
        ----------
        chainName : str
            The name of the chain from the collection.
        q : float, optional
            Quantiles of interest (in terms of percentiles, i.e. between 0 and
                                  1).
            The default is [0.025, 0.5, 0.975].
        r : float, optional
            Accuracy. The default is 0.005.
        s : float, optional
            Probability. The default is 0.95.
        converge_eps : float, optional
            Convergence threshold (epsilon). The default is 0.001.
        thin : int, optional
            Thining amount. The default is 1.
        print_each : bool, optional
            Print results at each evaluation. The default is False.
        print_final : bool, optional
            Print the final results. The default is False.

        Returns
        -------
        None.

        '''
        q = list(q)
        columns = ["Burn-in (M)", "Total (N)", "Lower Bound (Nmin)",
                   "Dependence Factor (I)", "Thinning (k)", "Quantile",
                   "Parameter"]
        all_samples = pd.DataFrame(columns=columns)
        for i in range(len(q)):
            needed_size = samp_size_calc_raftery(q[i], r, s)[1]
            if needed_size > len(self.chains[chainName]):
                print("not enough samples in the chain for quantile " +
                      str(q[i]) + ". Could not evaluate.")
            else:
                res = raftery(self.chains[chainName], q=q[i], r=r, s=s,
                              converge_eps=converge_eps, thin=thin,
                              print_=print_each)
                res["Quantile"] = q[i]
                res['Parameter'] = res.index
                all_samples = all_samples.append(res, ignore_index=True)

        maxcols = ["Max Burn-in (M)", "Max Total (N)",
                   "Max Lower Bound (Nmin)", "Max Dependence Factor (I)",
                   "Max Thinning (k)"]
        maxvals = np.max(all_samples.iloc[:, :-2])
        finalres = pd.DataFrame(np.array(maxvals).reshape(1, 5),
                                columns=maxcols)
        inputs = {"r": r, "s": s, "q": q}

        if print_final:
            print("\r\nInput Parameters: ", "\r\n",
                  "\tQuantile (q): ", inputs["q"], "\r\n",
                  "\tAccuracy (r): +/-", inputs["r"], "\r\n",
                  "\tProbability (s): ", inputs["s"], "\r\n",
                  "\r\n", finalres)
        self.diagnostic_results[chainName + "_Raftery"] = finalres

    def TJ_Convergence_test(self, chainName, eps=0.025, quantiles=[0.05, 0.95],
                            window_size=None, num_windows=5, slide=50,
                            window_space=0, bin_limit=0.6, print_final=False):
        '''
        A homemade test to evaluate convergence. This test evaluates a moving
        window, or a list of moving windows and compares the values of the
        distribution tails in those windows. If the distribution tails of all
        the moving windows is in line with the distribution tails of the final
        n samples of the chain, then the chain is considered to have converged.

        Parameters
        ----------
        chainName : str
            The name of the chain from the collection.
        eps : float, optional
            The threshold for comparing similarity in the chain. If the
            similarity between the moving window distributions and the
            distribution at the end of the chain is below the threshold, then
            the chain is considered to have converged. The default is 0.025.
        quantiles : list, or list-like, optional
            The sections of the distributions to consider for similarity.
            The default is [0.05, 0.95].
        window_size : int, optional
            How many chain samples to consider in each window. If None,
            it is automatically determined based on the size of the chain.
            The default is None.
        num_windows : int, optional
            The number of moving windows to use in the evaluation.
            The default is 5.
        slide : int, optional
            The number of samples to slide after each iteration, or in other 
            words, how fast the moving windows move (a larger value for slide
            means less total evaluations). The default is 50.
        window_space : int, optional
            The distance between each window. If positive, there is a gap. If
            negative, there is overlap. The default is 0.
        bin_limit : float, optional
            The bin_limit is the percent (between 0 and 1) of samples
            that are used in the moving windows, or in other words, the 
            evaluation stops once the right edge (i.e. most recently sampled 
            observation in the chain) matches the upper limit determined by
            the bin_limit percent. Also, 1 - bin_limit is the amount of samples
            from the end that are used as a baseline to see if the chain has
            converged. The default is 0.6.
        print_final : bool, optional
            Print the final results. The default is False.

        Raises
        ------
        ValueError
            Raises ValueError if argumnets passed by the user are outside
            function constraints.

        Returns
        -------
        None.

        '''
        chain = np.array(self.chains[chainName])
        # eps must be greater than 0
        # window size needs to be a number between 1 and
        # bin_limit of the length of the chain
        # bin_limit must be between 0 and 1
        if float(bin_limit) <= 0 or float(bin_limit) >= 1:
            raise ValueError("bin_limit must be between 0 and 1")
        if float(eps) <= 0:
            raise ValueError("""eps must be greater than 0 (its good to pick a
                                decimal close to zero, but not equal to it)""")
        if np.round(float(num_windows)) <= 0:
            raise ValueError("num_windows must be greater than 0")

        nrows = chain.shape[0]
        chain = chain.reshape(nrows, -1)
        ncols = chain.shape[1]

        if window_size is None:
            window_size = np.ceil(0.05*nrows)
        elif float(window_size) < 1 or float(slide) < 1:
            raise ValueError("window_size or slide must be at least 1")

        max_row = np.ceil(nrows * bin_limit)
        init_right_edge = (window_size * num_windows) + \
                          (window_space * (num_windows - 1))
        bins = np.floor((max_row - init_right_edge) / slide)

        col_names = ["Burn-in", "Ending Bin", "Total Bins", "Status",
                     "Notes", "Msg Details"]
        res = pd.DataFrame(columns=col_names)

        for col in range(ncols):
            cur_chain = chain[:, col]
            end_cur_chain = cur_chain[int(max_row):]
            quantiles.sort()
            end_per = np.percentile(end_cur_chain, 100*np.array(quantiles))
            bin_i = 1
            per_ratios = np.ones((num_windows, len(quantiles)))*(2+eps)
            bin_mat = np.arange(num_windows)

            if (num_windows == 1):
                index_vals = np.array([0.0])
            else:
                index_vals = (bin_mat * window_size) + (bin_mat * window_space)

            while (np.any(np.abs(per_ratios - 1) >= eps)
                    and bin_i <= bins):
                for i in range(len(index_vals)):
                    window_chain = cur_chain[
                            int(index_vals[i]):int(index_vals[i]+window_size)]
                    per_ratios[i, :] = np.percentile(
                            window_chain, 100*np.array(quantiles)) / end_per

                index_vals += slide
                bin_i += 1

            if bin_i > bins:
                msg = "Unsuccessful"
                sub_msg = """Fully iterated before meeting criteria, may not
                have stabilized on a distribution. Try adjusting the settings
                and try again, or take a look at the plot"""
            elif np.any(np.abs(per_ratios - 1) >= eps):
                msg = "Unsuccessful"
                sub_msg = """Did not appear to stabilize on a distribution.
                          Try adjusting the settings and try again, or take a
                          look at the plot"""
            else:
                msg = "Successful"
                sub_msg = ""
            msg_display = "See Msg Details in DataFrame"

            if num_windows % 2 == 1:
                burnin = index_vals[
                        int(np.median(bin_mat))] + (np.ceil(window_size / 2))
            else:
                burnin = index_vals[
                        int(np.floor(np.median(bin_mat)))] + window_size

            res = res.append(pd.DataFrame([[burnin, bin_i, bins, msg,
                                            msg_display, sub_msg]],
                                          columns=col_names),
                             ignore_index=True)

        inputs = {"eps": eps, "window_size": window_size,
                  "num_windows": num_windows, "slide": slide,
                  "window_space": window_space, "bin_limit": bin_limit}

        if print_final:
            print("\r\nInput Parameters: ", "\r\n",
                  "\tRatio Epsilon: ", inputs["eps"], "\r\n",
                  "\tWindow Size: ", inputs["window_size"], "\r\n",
                  "\tNumber of Windows: ", inputs["num_windows"], "\r\n",
                  "\tSlide Amount: ", inputs["slide"], "\r\n",
                  "\tSpace between Windows: ", inputs["window_space"], "\r\n",
                  "\tBin Limit Percentage: ", inputs["bin_limit"], "\r\n",
                  "\r\n", res.loc[:, res.columns != "Msg Details"])

        self.diagnostic_results[chainName + "_Convtest"] = res

    def runMCMC(self, start, initSampleSize, lpost, algo, algoOpts=None,
                raftOpts=None, chainName=None, max_tries=100, sd=0.02,
                plot_trace=True, plot_density=True, plot_acf=True,
                plot_trace_args=None, plot_density_args=None,
                plot_acf_args=None, acfType='pacf', acf_args=None,
                do_raftery=True, iters_to_go=None, max_iters=750000, burnin=0,
                lpost_args={}):
        '''
        TODO: Update to include ESS option, instead of just raftery.
        
        Generate MCMC samples and evaluate samples size (using Raftery)
        and convergence.
        
        Parameters
        ----------
        start : numpy array
            Starting values for the MCMC.
        initSampleSize : int
            The number of MCMC samples to draw on the first run. It's good to 
            start relatively small, because the Raftery evaluation will tell
            you how many more samples need to be drawn.
        lpost : function
            Log posterior function.
        algo : function
            The MCMC algorithm to use (could be anything, but needs to have the
            same arguments as inputs for the algorithms already defined, namely:
                start : numpy array
                    Starting values for the MCMC.
                niter : int
                    Number of iterations.
                lpost : function
                    Log posterior function.
                postArgs : dict
                    Extra arguments for the log posterior function. The default is
                    an empty dictionary
                options : dict, optional
                    Extra arguments for the specific MCMC algorithm
        algoOpts : dict, optional
            Extra arguments for the specific MCMC algorithm.
            The default is None.
        raftOpts : dict, optional
            A dictionary containing the options for the Raftery evaluation, 
            namely:
                q : float, optional
                    Quantiles of interest (in terms of percentiles, i.e.
                                           between 0 and 1).
                    The default is [0.025, 0.5, 0.975].
                r : float, optional
                    Accuracy. The default is 0.005.
                s : float, optional
                    Probability. The default is 0.95.
                converge_eps : float, optional
                    Convergence threshold (epsilon). The default is 0.001.
                thin : int, optional
                    Thining amount. The default is 1.
                print_each : bool, optional
                    Print results at each evaluation. The default is False.
                print_final : bool, optional
                    Print the final results. The default is False.
            The default is None.
        chainName : str, optional
            The name of the chain that will be created. If None,
            'Chain_' + int (for number of chains in the collection) 
            will be used. The default is None.
        max_tries : int, optional
            The max number of times to try and jitter before admitting defeat.
            If the jitter fails, the reason or the covaraince matrix not being
            positive definite may not be due to randomness, and may require
            a re-evaluation of the problem space. The default is 100.
        sd : float, optional
            The standard deviation of the normal distribution used to draw the
            jitter amount from. In other words, the jittered covariance is the 
            covaraince matrix plus a random draw X, where X~N(0, sd). 
            The default is 0.02.
        plot_trace : bool, optional
            Plot the trace of the MCMC samples. The default is True.
        plot_density : bool, optional
            Plot the posterior density of the MCMC samples.
            The default is True.
        plot_acf : bool, optional
            Plot the auto-correlation. The default is True.
        plot_trace_args : dict
            Arguments for the plotTrace function. Default is None.
        plot_density_args : dict
            Arguments for the plotDensity function. Default is None.
        plot_acf_args : dict
            Arguments for the plotACF function. Default is None.
        acf_type : str
            Ether 'acf', 'pacf', or None. Default is 'pacf'. If None, pacf
            calculation is not performed.
        acf_args : dict
            Arguments to pass to the `acf` function. Default is None.
        do_raftery : bool, optional
            Whether to perform the raftery evaluation, or stop after the
            first chain generation. Default is True.
        max_iters : int, optional
            The max number of new samples to draw. For example, if the Raftery
            evaluation recommends 1 million new samples, and max_iters is 
            750000, then the new samples will be restricted at 750000.
            The default is 750000.
        burnin : int, optional
            The number of initial MCMC samples to drop from the chain.
            If burnin is negative or zero, the burnin amount will be determined
            automatically. If positive, it will drop that amount. If 
            None, it will do nothing. The default is 0.
        lpost_args : dict, optional
            Any extra arguments to pass to the log posterior function.
            The default is an empty dictionary.

        Returns
        -------
        None.

        '''

        if algoOpts is None:
            algoOpts = {}
        first_run_results = applyMCMC(st=start, ni=initSampleSize, lp=lpost,
                                      algo=algo, algoOpts=algoOpts,
                                      postArgs=lpost_args,
                                      sd=sd, max_tries=max_tries)
        # TODO: change this so it's not hardcoded
        previous_values = first_run_results['prev_vals']
        new_start = first_run_results['parameters'][-1]

        self.addChain(first_run_results['parameters'], chainName, concat=False)
        self.previous_values[chainName + "_latestrun"] = previous_values

        if do_raftery:
            qq = [0.025, 0.5, 0.975]
            rr = 0.005
            ss = 0.90
            ce = 0.001
            th = 1
            pe = False
            pf = False
            if raftOpts is not None:
                keys = list(raftOpts.keys())
                if 'q' in keys:
                    qq = raftOpts['q']
                if 'r' in keys:
                    rr = raftOpts['r']
                if 's' in keys:
                    ss = raftOpts['s']
                if 'converge_eps' in keys:
                    ce = raftOpts['converge_eps']
                if 'thin' in keys:
                    th = raftOpts['thin']
                if 'print_each' in keys:
                    pe = raftOpts['print_each']
                if 'print_final' in keys:
                    pf = raftOpts['print_final']
            
            self.bestRaftery(chainName, q=qq, r=rr, s=ss, converge_eps=ce,
                             thin=th, print_each=pe, print_final=pf)
            raftmin = int(self.diagnostic_results[chainName + "_Raftery"]["Max Total (N)"].values)
            initial_len = self.chains[chainName].shape[0]
            iters_to_go = raftmin - initial_len
            
            if iters_to_go >= 1:
                if iters_to_go >= max_iters:
                    iters_to_go = max_iters
                pv = self.previous_values[chainName + "_latestrun"]
                algoOpts.update({'prev_vals': pv})
                final_run_results = applyMCMC(st=new_start,
                                              ni=int(iters_to_go),
                                              lp=lpost,
                                              algo=algo,
                                              algoOpts=algoOpts,
                                              postArgs=lpost_args,
                                              sd=sd,
                                              max_tries=max_tries)
                previous_values = final_run_results['prev_vals']
                self.addChain(final_run_results['parameters'], chainName, concat=True)
                self.previous_values[chainName + "_latestrun"] = previous_values

        burnin_param = None
        burnVal = 0
        if burnin is not None:
            if burnin <= 0:
                self.TJ_Convergence_test(chainName)
                conv_diag = self.diagnostic_results[chainName + "_Convtest"]
                # burnin_param gets used in plotting later. It's possible for each
                # parameter to have a seperate optimal burn-in point, however,
                # the parameters themselves should not be considered 
                # independently, and thus, need a common burn-in value
                burnin_param = np.array(conv_diag["Burn-in"])
                burnVal = int(conv_diag["Burn-in"].max())
            else:
                # burnin_param is for plotting purposes only
                burnin_param = np.repeat(burnin, len(start))
                burnVal = burnin

        if plot_trace:
            if plot_trace_args is None:
                plot_trace_args = {'CTres': burnin_param,
                                   'write': False,
                                   'pdir': "./Plots/",
                                   'fileType': "png",
                                   'figsize': (15, 12)
                                   }
            else:
                plot_trace_args.update({'CTres': burnin_param})
            self.plotTrace(chainName, **plot_trace_args)
        if plot_density:
            if plot_density_args is None:
                plot_density_args = {'smoothing': 0.05,
                                     'write': False,
                                     'pdir': "./Plots/",
                                     'vlines': None,
                                     'fileType': "png",
                                     'figsize': (15, 12)
                                     }
            self.plotDensity(chainName, **plot_density_args)
        if acfType is not None:
            if acfType == 'pacf':
                partial = True
            else:
                partial = False
            
            if acf_args is None:
                acf_args = {'lag': 50,
                            'partial': partial,
                            'demean': True}
            else:
                acf_args.update({'partial': partial})
            
            self.acf(chainName, **acf_args)
            if plot_acf:
                if plot_acf_args is None:
                    plot_acf_args = {'bounds': True,
                                     'ci': 0.95,
                                     'acfType': acfType,
                                     'write': False,
                                     'pdir': "./Plots/",
                                     'fileType': "png",
                                     'lw': None,
                                     'figsize': (15, 12)
                                     }
                self.plotACF(chainName, **plot_acf_args)

        self.burnin(chainName, burnVal)

    def plotTrace(self, chainName, CTres=None, write=False,
                  display=True, pdir="./Plots/",
                  fileType="png", figsize=(15, 12)):
        '''
        Plot the trace of the MCMC chain.

        Parameters
        ----------
        chainName : str
            The name of the MCMC chain.
        CTres : numpy array, optional
            The results of the TJ_Convergence_Test. The default is None.
        write : bool, optional
            Write plot to a directory. The default is False.
        display : bool, optional
            Display the plot.
        pdir : str, optional
            The directory to write the plots to. The default is "./Plots/".
        fileType : str, optional
            The filetype of the image. The default is "png".
        figsize : tuple, optional
            The figure size (see matplotlib documentation for more details).
            The default is (15, 12).

        Returns
        -------
        fig, ax : tuple
            The figure and axes components of the plot.

        '''
        trace = self.chains[chainName]
        trace = trace.reshape(trace.shape[0], -1)
        nparam = trace.shape[1]
        fig, ax = plt.subplots(nrows=nparam, ncols=1, figsize=figsize,
                               squeeze=False)
        for i in range(nparam):
            ax[i, 0].plot(trace[:, i], label='Sample Values')
            ax[i, 0].set_ylabel(''.join(["Value for Parameter ",
                                         str(i+1), "/",
                                         str(nparam), " Value"]))
            ax[i, 0].set_xlabel("Iteration (Sample) Number")
            if i == 0:
                ax[i, 0].set_title("Trace Plot for " + self.name +
                                   " Parameters")
            if CTres is not None:
                ax[i, 0].axvline(x=int(CTres[i]), color="red",
                                 linewidth=2.0, label='Recommended Burnin')
            ax[i, 0].legend()
        if write:
            pathlib.Path(pdir).mkdir(exist_ok=True)
            fig.savefig(''.join([pdir, self.name, '_', chainName,
                                 "_trace.", fileType]))
        if display:
            fig.show()
        #plt.close()
        return fig, ax

    def plotDensity(self, chainName, smoothing=0.05, write=False,
                    display=True, pdir="./Plots/", vlines=None,
                    fileType="png", figsize=(15, 12)):
        '''
        Plot the density of the MCMC chain.

        Parameters
        ----------
        chainName : str
            The name of the MCMC chain.
        smoothing : float, optional
            The amount of smoothing to use on the kde plot.
            See seaborn.kde_plot for details.
            The default is 0.05.
        write : bool, optional
            Write plot to a directory. The default is False.
        display : bool, optional
            Display the plot.
        pdir : str, optional
            The directory to write the plots to. The default is "./Plots/".
        vlines : TYPE, optional
            The x-axis locations of any predetermined vertical lines on the
            density plots, such as mean, median, or mode.
            The default is None.
        fileType : str, optional
            The filetype of the image. The default is "png".
        figsize : tuple, optional
            The figure size (see matplotlib documentation for more details).
            The default is (15, 12).

        Returns
        -------
        fig, ax : tuple
            The figure and axes components of the plot.

        '''
        trace = self.chains[chainName]
        trace = trace.reshape(trace.shape[0], -1)
        nparam = trace.shape[1]
        fig, ax = plt.subplots(nrows=1, ncols=nparam, figsize=figsize,
                               sharey=True, squeeze=False)
        for i in range(nparam):
            sns.kdeplot(trace[:, i], ax=ax[0, i], shade=True)
            if i == 0:
                ax[0, i].set_ylabel("Density")
            ax[0, i].set_xlabel(''.join(["Value for Parameter ", 
                                         str(i+1), "/", str(nparam)]))
            if vlines is not None:
                ax[0, i].axvline(vlines[i])
        fig.suptitle("Posterior Density of " + self.name + " Parameters")
        if write:
            pathlib.Path(pdir).mkdir(exist_ok=True)
            fig.savefig(''.join([pdir, self.name, '_', chainName, "_density.",
                                 fileType]))
        if display:
            fig.show()
        #plt.close()
        return fig, ax

    def plotACF(self, chainName, bounds=True, ci=0.95, acfType="acf",
                write=False, display=True, pdir="./Plots/", fileType="png",
                lw=None, figsize=(15, 12)):
        '''
        Plot the Autocorrelation function of the chain.

        Parameters
        ----------
        chainName : str
            The name of the MCMC chain.
        bounds : bool, optional
            Draw the bounds of the autocorrelation. The default is True.
        ci : float, optional
            The size of the bounds (confidence interval), if applicable.
            The default is 0.95.
        acfType : str, optional
            The type of acf plot to draw. Can be either 'acf' or 'pacf'.
            The default is "acf".
        write : bool, optional
            Write plot to a directory. The default is False.
        display : bool, optional
            Display the plot.
        pdir : str, optional
            The directory to write the plots to. The default is "./Plots/".
        fileType : str, optional
            The filetype of the image. The default is "png".
        lw : float, optional
            The line width to use on the plot. If None, it will be calculated
            automatically. The default is None.
        figsize : tuple, optional
            The figure size (see matplotlib documentation for more details).
            The default is (15, 12).

        Returns
        -------
        fig, ax : tuple
            The figure and axes components of the plot.

        '''
        try:
            self.diagnostic_results[chainName + "_" + acfType]
        except KeyError:
            print("""No ACF found. Please calculate ACF, pACF,
                  or a variant using available methods""")
        allacf = self.diagnostic_results[chainName + "_" + acfType]
        allacf = allacf.reshape(allacf.shape[0], -1)
        (samples, nparam) = allacf.shape
        fig, ax = plt.subplots(nrows=nparam, ncols=1, figsize=figsize,
                               squeeze=False)
        # Picked this as the line width value because it seems to pick a good
        # width with respect to the lag number. the plot is designed to mimic
        # the R acf() function
        if lw is None:
            lw = 1-np.exp(-0.00346103*(samples-1))
        for i in range(nparam):
            ax[i, 0].bar(range(len(allacf[:, i])), allacf[:, i], width=lw)
            ax[i, 0].set_ylabel(''.join([acfType.upper(), " for Param. ",
                                         str(i+1), "/", str(nparam)]))
            ax[i, 0].axhline(y=0, linewidth=0.5)
            if i == 0:
                ax[i, 0].set_title("ACF Plot for " + self.name +
                                   " Parameters")
            if i == nparam - 1:
                ax[i, 0].set_xlabel("Lag")
            if bounds:
                bnd = qnorm_aprox((1+ci)/2)/np.sqrt(samples)
                ax[i, 0].axhline(y=bnd, color="red",
                                 linestyle="dashed", linewidth=0.5)
                ax[i, 0].axhline(y=-bnd, color="red",
                                 linestyle="dashed", linewidth=0.5)
        if write:
            pathlib.Path(pdir).mkdir(exist_ok=True)
            fig.savefig(''.join([pdir, self.name, '_', chainName, acfType, ".", fileType]))
        if display:
            fig.show()
        #plt.close()
        return fig, ax

    def acf(self, chainName, lag=50, partial=False, demean=True):
        '''
        ACF definition for a wide-sense stationary process, partial acf uses
        Yule-Walker MLE method.

        Parameters
        ----------
        chainName : str
            The name of the MCMC chain.
        lag : int, optional
            The lag in the autocorrelation. The default is 50.
        partial : bool, optional
            Calculate pACF instead of ACF. The default is False.
        demean : bool, optional
            Center the chain before calculating autocorrelation.
            The default is True.

        Returns
        -------
        None.

        '''
        trace = self.chains[chainName]
        trace = trace.reshape(trace.shape[0], -1)
        params = trace.shape[1]
        samples = trace.shape[0]
        allacf = np.zeros((lag + 1, params))
        for param in range(params):
            x = trace[:, param]
            if demean:
                center = x - x.mean()
            else:
                center = x
            acf = np.zeros(lag + 1)
            acf[0] = 1
            z = np.sum(center**2)/samples ## variance of x
            for l in range(1, lag+1):
                if partial:
                    r = np.zeros(l+1)
                    r[0] = z
                    for k in range(1, l+1):
                        headlag = center[k:]
                        taillag = center[:-k]
                        r[k] = np.sum(headlag*taillag)/(samples-k)
                    r[l] = np.sum(headlag*taillag)/(samples-l)
                    R = toeplitz(r[:-1])
                    rho = solve(R, r[1:])
                    #sigma2 = r[0] - (r[1:]*rho).sum()
                    acf[l] = rho[-1]
                else:
                    headlag = center[l:]
                    taillag = center[:-l]
                    acf[l] = np.sum(headlag*taillag)/samples/z
            allacf[:, param] = acf
        if partial:
            self.diagnostic_results[chainName + "_pacf"] = allacf
        else:
            self.diagnostic_results[chainName + "_acf"] = allacf
                    


import os
import codecs

from setuptools import setup, find_packages


with open("README.md", 'r', encoding='utf-8') as f:
    long_description = f.read()

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()
    
def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")
    
setup(
    name='tsdst',
    version=get_version("tsdst/__init__.py"),
    description='A low-key data science and statistics toolkit',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Tom W',
    author_email='tomwerner5@gmail.com',
    url='https://tomwerner5.github.io/tsdst/',
    download_url='https://github.com/tomwerner5/tsdst/archive/v_' + get_version("tsdst/__init__.py") + 'tar.gz',
    packages=find_packages(include=['tsdst', 'tsdst.nn']),
    keywords=['data science', 'statistics', 'neural network', 'bayesian'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Mathematics'
    ],
    install_requires=['matplotlib>=3.1.0',
                      'numba>=0.44.1',
		      'numdifftools>=0.9.39',
		      'numpy>=1.16.4',
                      'pandas>=0.24.2',
                      'scikit-learn>=0.22.1',
                      'scipy>=1.2.1',
		      'seaborn>=0.11.0',
                      'sqlalchemy>=1.3.4',
                      'statsmodels>=0.9.0',
                      ]
)
    



from __future__ import (division, generators, absolute_import,
                        print_function, with_statement, nested_scopes,
                        unicode_literals)

import numpy as np
import pandas as pd
import warnings

from itertools import chain, combinations
from numpy.linalg import inv
from scipy.stats import rankdata, kendalltau, gaussian_kde


def custom_abs(x):
    '''
    Innards of the absolute value function. (This is just here for fun).
    
    Parameters
    ----------
    x : signed float or int
        The number to be evaluated.

    Returns
    -------
    int or float
        The absolute value of x.

    '''
    m = x % (x**2 - x + 2)
    p = m % (x**2 + 2)
    return 2*p - x


def powerset(iterable):
    '''
    Create the powerset (all possible combinations) of a list.

    Parameters
    ----------
    iterable : list or list-like
        The collection of items to create a powerset from.

    Returns
    -------
    list of lists
        list containing all possible combinations.

    '''
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))
  

def percentIncrease(old, new):
    '''
    Calculates the percent increase between two numbers.

    Parameters
    ----------
    old : numeric
        The old number.
    new : numeric
        The new number.

    Returns
    -------
    numeric
        The percent increase (as a decimal).

    '''
    return (new - old)/old


# This will do a 'first' ranking of the data and return the ranks.
# For R-like rank functionality, import scipy's rankdata from 
# this module
def rank(x, small_rank_is_high_num=True, rank_from_1=True):
    '''
    Rank items in an array. Using the 'first' method, which ranks ties using
    the order of appearance. For rank functionality similar to R, see scipy's
    rankdata function (which is imported from this module for convenience).'

    Parameters
    ----------
    x : numpy array or array-like
        The array of values to be sorted.
    small_rank_is_high_num : bool, optional
        Smallest rank value is the highest/largest number.
        The default is True.
    rank_from_1 : bool, optional
        Use 1 as the top rank rather than 0. The default is True.

    Returns
    -------
    rk : numpy array
        An array containing the ranks.

    '''
    rk = np.argsort(np.argsort(x))
    if small_rank_is_high_num:
        rk = (len(x) - 1) - rk
    if rank_from_1:
        rk += 1
    return rk


def cov2cor(V):
    '''
    Translation of R's cov2cor function

    Parameters
    ----------
    V : numpy array
        Covariance matrix.

    Returns
    -------
    R : numpy array
        Correlation matrix.

    '''
    p, n = V.shape
    if len(V.shape) != 2 or p != n:
        raise ValueError('V is not a square matrix')
    Is = np.sqrt(1/np.diag(V))
    if np.any(~np.isfinite(Is)):
        warnings.warn("""Diagonal had 0, NA, or infinite entries. Non-finite
                      result is doubtful""")
    r = Is.reshape(-1, 1) * V * np.repeat(Is, p).reshape(p, p).T
    return r
    

def mann_whitney(x, y, y_is='groups'):
    '''
    Mann-Whitney sum-rank test.

    Parameters
    ----------
    x : numpy array
        Values to be evaluated. Either a single group, or an array containing
        both group values if y_is == 'groups'
    y : numpy array
        Values to be evaluated against x, or, if y_is == 'groups',
        then y is a boolean value indicating which group the values
        of x belong to.
    y_is : bool, optional
        Whether or not y is itself a group of values, or an indicator variable
        for identifying the groups within x. The default is 'groups'.

    Returns
    -------
    res : dict
        The results of the Mann-Whitney test.

    '''
    if isinstance(y, (list, tuple)):
        y = np.array(y)
    if isinstance(x, (pd.DataFrame, pd.Series)):
        x = x.values
    if y_is == 'groups':
        X = x
        Y = y
        overall_rank = rankdata(X, 'average')
    else:
        if len(x.shape) >= 2:
            x = x.ravel()
        if len(y.shape) >= 2:
            y = y.ravel()
        X = np.concatenate((x, y))
        Y = np.concatenate((np.repeat(0, x.size), np.repeat(1, y.size)))
        overall_rank = rankdata(X, 'average')
    uni_groups = np.unique(Y)
    rank_group1 = overall_rank[Y == uni_groups[0]]
    rank_group2 = overall_rank[Y == uni_groups[1]]
    sum_ranks_g1 = rank_group1.sum()
    sum_ranks_g2 = rank_group2.sum()
    
    n1 = rank_group1.shape[0]
    n2 = rank_group2.shape[0]
    
    U1 = sum_ranks_g1 - (n1*(n1 + 1))/2
    U2 = sum_ranks_g2 - (n2*(n2 + 1))/2
    
    res = {'U1': U1,
           'U2': U2,
           'n1': n1,
           'n2': n2,
           'rank_group1': rank_group1,
           'rank_group2': rank_group2,
           'sum_ranks_g1': sum_ranks_g1,
           'sum_ranks_g2': sum_ranks_g2}
    
    res['max_U'] = 'U1' if U1 >= U2 else 'U2'
    return res


def biserial_rank_corr(x, y, y_is='groups'):
    '''
    Biserial Rank Correlation.

    Parameters
    ----------
    x : numpy array
        Values to be evaluated. Either a single group, or an array containing
        both group values if y_is == 'groups'
    y : numpy array
        Values to be evaluated against x, or, if y_is == 'groups',
        then y is a boolean value indicating which group the values
        of x belong to.
    y_is : bool, optional
        Whether or not y is itself a group of values, or an indicator variable
        for identifying the groups within x. The default is 'groups'.

    Returns
    -------
    r : float
        Biserial rank correlation.

    '''
    mw = mann_whitney(x, y, y_is=y_is)
    max_u = mw[mw['max_U']]
    r = (2*max_u)/(mw['n1']*mw['n2']) - 1
    return r


# mimics R cor function. Columns are correlation variables.
# X and Y must at least have the same length
def corr(x, y=None, method='pearson'):
    '''
    Mimics the R cor function to calculate correlations.

    Parameters
    ----------
    x : numpy array or pandas dataframe
        Either a matrix (to calculate correlations across columns) or a 1D
        array to compare with y.
    y : numpy array or pandas series, optional
        A 1D array to compare with a 1D x, or None if x is a matrix.
        The default is None.
    method : str, optional
        Type of correlation to compute. The default is 'pearson'.

    Raises
    ------
    ValueError
        Raised if method provided is invalid.

    Returns
    -------
    numpy array
        Array containing the correlation.

    '''
    cor_mat = None
    X = np.array(x, ndmin=2)
    if X.shape[0] == 0:
        return np.array([]).reshape(0, 0)
    if X.shape[0] == 1:
        X = X.T
    if y is not None:
        Y = np.array(y, ndmin=2)
        if Y.shape[0] == 1:
            Y = Y.T
        X = np.concatenate((X, Y), axis=1)
    if method == 'pearson':
        cor_mat = np.corrcoef(X, rowvar=False)
    elif method == 'spearman':
        M = np.apply_along_axis(rankdata, 0, X)
        D = np.eye(X.shape[1])*M.std(ddof=1, axis=0)
        Dinv = inv(D)
        cor_mat = Dinv.dot(np.cov(M, rowvar=False)).dot(Dinv)
    elif method == 'kendall':
        cor_mat = np.zeros((X.shape[1], X.shape[1]))
        corrs = []
        for i in range(X.shape[1]):
            for j in range(i, X.shape[1]):
                c, _ = kendalltau(X[:, i], X[:, j])
                corrs.append(c)
        corrs = np.array(corrs)
        row, col = np.triu_indices(X.shape[1])
        cor_mat[row, col] = corrs
        cor_mat[col, row] = corrs
    else:
        raise ValueError('Not a valid correlation method. (%s) ' % method)
    
    if isinstance(x, pd.DataFrame):
        cor_mat = pd.DataFrame(cor_mat, index=x.columns, columns=x.columns)
    return cor_mat


def mode_histogram(data, bin_max='auto', edge_check=None, approx=False,
                   surround=True, surround_step=1):
    # TODO: make this work in a multivariate setting
    '''
    Calculates the mode (from a continuous distribution) using data provided
    from a histogram. Created because calculating mode from KDE estimate
    can be slow.

    Parameters
    ----------
    data : numpy array
        Array of data to calculate the mode from.
    bin_max : int or str, optional
        Number of histogram bins to pass to np.histogram. Can be string or
        integer. The default is 'auto'.
    edge_check : str or None, optional
        The edge of the largest histogram rectangle (mode) to use for
        comparing to the original data. For example, if the mode
        rectangle of the histogram created is between 1 and 3 on the x-axis,
        then 'left' uses 1 for comparison, 'right' uses 3, and 'center' uses
        the midpoint 2. If None, the optimal metod is used based on the number
        of bins. Default is None.
    approx : bool, optional
        If approx is True, use the histogram output as the mode. If False, find
        the closest value in data to the histogram value and use that as the
        mode.

    Returns
    -------
    float
        The mode of the array.

    '''
    if isinstance(bin_max, str):
        bins = bin_max
    else:
        if len(data) < bin_max:
            bins = len(data)
        else:
            bins = bin_max
    hist, edges = np.histogram(data, bins=bins)
    hist_max_idx = np.argmax(hist)
    if edge_check is None:
        if hist.shape[0] % 2 == 0:
            edge_check = "left"
        else:
            edge_check = "center"
    
    if surround:
        hist_max_indices = np.arange(hist_max_idx - surround_step, hist_max_idx + surround_step + 1, 1)
        if hist_max_indices[0] < 0:
            hist_max_indices = hist_max_idx
        hist_max_values = hist[hist_max_indices]
    else:
        hist_max_indices = hist_max_idx
    
    if edge_check == "center" or edge_check == "centre":
        max_edge = edges[hist_max_indices] + (edges[hist_max_idx + 1] - edges[hist_max_idx])/2
    elif edge_check == "left":
        max_edge = edges[hist_max_indices]
    elif edge_check == "right":
        max_edge = edges[hist_max_indices + 1]
    else:
        raise ValueError("Not a valid edge_check. See help(mode_histogram)")
    
    if surround:
        max_edge = np.sum(max_edge*(hist_max_values/hist_max_values.sum()))
    
    if approx:
        mode = max_edge
    else:
        mode_idx = np.abs(np.asarray(data) - max_edge).argmin()
        mode = data[mode_idx]
    return mode
    

def mode_kde(data, kde_args=None):
    '''
    Calculate mode from gaussian kernel density estimate (using scipy
    gaussian_pde).

    Parameters
    ----------
    data : numpy array
        Array to calculate mode from.
    kde_args : dict
        A dictionary containing arguments for gaussian_kde (scipy)

    Returns
    -------
    mode : 
        A numeric value, which is the mode.

    '''
    if kde_args is None:
        kde_args = {}
    kernel = gaussian_kde(data, **kde_args)
    height = kernel.pdf(data)
    mode = data[np.argmax(height)]
    return mode


def norm(x, p, use_expo=False):
    '''
    Calculate LP norm of a vector.

    Parameters
    ----------
    data : numpy array
        A 1D numpy array of values.
    p : int
        The LP norm specification (1,2,3, ...).

    Returns
    -------
    numeric
        Value of LP norm on the vector.

    '''
    expo = 1/p
    if p == 1:
        return np.sum(np.abs(x))
    elif p >= 2:
        if p % 2 == 0:
            norm = np.sum(x**p)
        else:
            norm = np.sum(np.abs(x)**p)
        # Regression applications with L2 penalty often dismiss the full norm,
        # particularily in L2 regularization. I'll note here that there is no added benefit
        # to using norms higher than 2 in those cases. Odd norms (L3 and higher) are non-differntiable at the origin
        # and they behave the same as L1. L4 and higher norms have the same 
        # local meaning are differentiable, thus providing no additional useful
        # information in regularization.
        if use_expo:
            return norm**expo
        else:
            return norm
    elif p == 0:
        # Note: Hastie defines the L_0 norm as the number of elements
        # though this is not derived mathmatically
        return len(x)
    else:
        raise('Not a valid norm')
        
        
def norm_der(x, p, use_expo=False):
    norm_der = (p*np.abs(x)**(p-1))*np.sign(x)
    if use_expo:
        norm_der = norm_der*((1/p)*np.sum(np.abs(x)**p)**(1/p - 1))
    return norm_der

        
def mahalanobis(data, produce=None):
    '''
    Calculate mahalanobis distance on a matrix of column vectors. 
    Assumes that rows are observations and columns are features.
    
    Parameters
    ----------
    data : numpy array or pandas dataframe
        The data to calculate distances on (columns are variables, rows are
        observations).
    produce : str, optional
        Variation of the output to produce, either `squared`, `leverage',
        or `sqrt` (None). The default is None.

    Returns
    -------
    numpy array
        Array containing the distances.

    '''
    arr = np.array(data).reshape(data.shape[0], -1)
    cent = arr - arr.mean(axis=0)
    covmat = np.cov(cent, rowvar=False)
    invcov = None
    if arr.shape[1] == 1:
        invcov = 1/covmat
    else:
        try:
            invcov = np.linalg.inv(covmat)
        except np.linalg.LinAlgError:
            invcov = np.linalg.pinv(covmat)
    md2 = np.sum(cent.dot(invcov) * cent, axis=1)
    if produce == "squared":
        return md2
    elif produce == "leverage":
        n = data.shape[0]
        return ((md2/(n - 1)) + (1/n))
    else:
        return np.sqrt(md2)


def scaling(x, a=0, b=1, percentile=(0, 100),
            center=None, scale=None, ddof=0):
    '''
    Common scaling algorithm. Either min-max (with any desired bound),
    standardizing, max absolute scaling, or robust scaling.

    Parameters
    ----------
    x : numpy array
        The data.
    a : float, optional
        The minimum value desired. The default is 0.
    b : float, optional
        The maximum value desired. The default is 1.
    percentile : list or array-like, optional
        Provide percentile bounds on values. The default is (0, 100).
    center : float or str, optional
        Either a value representing the centrality measure, or a string
        for the measure to be computed, either mean or median.
        If None, use the minimum of the percentile bounds.
        The default is None.
    scale : float or str, optional
        Either a value representing the dispersion measure, or a string
        for the measure to be computed, either std (standard deviation),
        var (variance), or maxAbs.
        If None, use the absolute difference of the percentile bounds.
        The default is None.
    ddof : int, optional
        Degrees of freedom for the variance/std. dev measure. The default is 0.

    Raises
    ------
    TypeError
        Raised if center is an invalid type.

    Returns
    -------
    scaled : numpy array
        The scaled values.

    '''
    lower = upper = centVal = scaleVal = None
    if center is not None:
        try:
            centVal = x - center
        except TypeError:
            if center == "mean":
                centVal = x - np.nanmean(x)
            elif center == "median":
                centVal = x - np.nanmedian(x)
            else:
                raise TypeError("""center must be a string, either 'mean' or
                                'median', else a number (central tendency)""")
    else:
        (lower, upper) = np.nanpercentile(x, percentile)
        centVal = x - lower

    if scale is not None:
        try:
            scaleVal = scale/1
        except TypeError:
            if scale == "maxAbs":
                scaleVal = np.nanmax(np.abs(x))
            elif scale == "std":
                scaleVal = np.nanstd(x, ddof=ddof)
            elif scale == "var":
                scaleVal = np.nanvar(x, ddof=ddof)
    else:
        if lower is None:
            (lower, upper) = np.nanpercentile(x, percentile)
        scaleVal = upper - lower

    scaled = (b - a)*centVal/scaleVal + a

    return scaled
        


import numpy as np
from scipy.special import xlogy


def relu(x, alpha=0):
    z = x.copy()
    # If alpha is between 0 and 1, the function performs leaky relu
    # alpha values are commonly between 0.1 and 0.3 for leaky relu
    z[z < 0] = z[z < 0]*alpha
    return z


def relu_der(x, alpha=0):
    # Note: relu derivative is technically undefined at x=0, but
    # tensorflow.nn.relu uses 0 for this special case, so that's what
    # I do here
    
    # If alpha != 0, then derivative is leaky relu
    dZ = x.copy()
    dZ[x <= 0] = alpha
    dZ[x > 0] = 1
    return dZ


def elu(x, alpha=0.1):
    '''
    Exponential Linear Unit.
    
    Main pro is it avoids dead relu. main con is computation time.
    
    Parameters
    ----------
    x : numpy array
        Values to be activated.
    alpha : float, optional
        The scale factor for the linear unit.
        Typical values are between 0.1 and 0.3.
        The default is 0.1.

    Returns
    -------
    z : numpy array
        The activated values.

    '''
    z = x.copy()
    
    z[z < 0] = alpha*(np.exp(z[z < 0]) - 1)
    return z


def elu_der(x, alpha=0.1):
    dZ = x.copy()
    dZ[x <= 0] = alpha*(np.exp(dZ[x <= 0]) - 1) + alpha
    dZ[x > 0] = 1
    return dZ


def sigmoid(x):
    return 1/(1 + np.exp(-x))


def sigmoid_der(x):
    sigm = sigmoid(x)
    der = sigm * (1 - sigm)
    return der


def softmax(x, center=True):
    # for numerical stability only
    x_temp = x.copy()
    if center:
        # TODO: see which is faster, axis=1, or global max
        x_temp -= np.max(x, axis=1, keepdims=True)
        #x_temp = (x_temp - np.mean(x))/np.std(x_temp)
    expx = np.exp(x_temp)
    return expx/(np.sum(expx, axis=1, keepdims=True))


def softmax_oneout(x, center=True):
    # for numerical stability only
    if center:
        x_temp = x - np.max(x)
    else:
        x_temp = x.copy()
    x_temp = x.copy()
    x_temp[:, x_temp.shape[1]-1] = 0
    expx = np.exp(x_temp)
    return expx/(np.sum(expx, axis=1, keepdims=True))


def create_jac_mask(shape, n_classes):
    mask = np.zeros(shape=shape)
    
    for i in range(int(shape[0]/n_classes)):
        mask[(i*n_classes):((i+1)*n_classes),
             (i*n_classes):((i+1)*n_classes)] = 1
    return mask


def softmax_der(x, center=True):
    # full jacobian, but diagonal is the only needed result
    soft = softmax(x, center=True).reshape(-1, 1)
    kronecker = np.dot(soft, soft.T)
    jac_mask = create_jac_mask(shape=kronecker.shape, n_classes=x.shape[1])
    
    out = np.diagflat(soft) - kronecker*jac_mask
    return out


def selu(x, alpha=1.6732632423543772848170429916717,
         lamda=1.0507009873554804934193349852946):
    # scaled exponential relu
    # Note: If lambda=1, then selu reduces to elu
    # Note: vanishing/exploding gradients are impossible with selu
    dZ = x.copy()
    dZ[x > 0] = lamda
    dZ[x <= 0] = lamda*(alpha*np.exp(x[x <= 0]))
    return dZ


def selu_der(x, alpha=1.6732632423543772848170429916717,
             lamda=1.0507009873554804934193349852946):
    # scaled exponential relu
    # Note: If lambda=1, then selu reduces to elu
    z = x.copy()
    z[z > 0] = lamda*z[z > 0]
    z[z <= 0] = lamda*(alpha*(np.exp(z[z < 0]) - 1))
    return z


def cross_entropy_binary(y_true, y_pred, delta=1e-9):
    '''
    Binary Cross Entropy.
    
    While the definition varies a little bit across ML or information theory
    domains, Cross-entropy in general is a method of measuring the difference
    of information between two probability distributions (i.e. a way of
    measuring how similar they are). In our context, cross entropy is the
    difference between our observed values (which can be viewed as a
    probability distribution where every probability is either 0 or 1, since
    every value is known and thus has an extreme probability) and the
    predicted values (which are actual probabilities since the true values
    are unknown). Since we are interested in predicting values that follow a
    bernoulli distribution, the cross entropy takes the form of the negative
    log-likelihood of the bernoulli distribution.
    
    With this cost function defined, Neural Networks are just performing
    a more BA version of maximum likelihood estimation. A negative cost is
    defined because maximizing the likelihood is the same as minimizing the
    negative likelihood. 
    
    Parameters
    ----------
    y_true : numpy array
        True, observed values. The outcome of an event (where 1 == success).
    y_pred : numpy array
        The predicted success probabilities.
    delta : float, optional
        A small, positive constant to offset predicted probabilities that are
        zero, which avoids log(0). Is ignored if delta = 0.
        The default is 1e-9.

    Returns
    -------
    cost : float
        The binary cross-entropy.

    '''

    # Compute the cross-entropy cost
    # To avoid log(0) errors (not necessary in most cases)
    ypred = y_pred.copy()
    if delta != 0:
        ypred[ypred <= delta] = delta
        ypred[ypred >= 1-delta] = 1-delta
    
    # m is the number of observations, and m_scale is a scaling factor to make
    # the computation easier in case the gradients are really big. 
    cost = -np.sum(xlogy(y_true, ypred) + xlogy(1 - y_true, 1 - ypred))
    return cost


def cross_entropy(y_true, y_pred, delta=1e-9):
    '''
    Binary Cross Entropy.
    
    While the definition varies a little bit across ML or information theory
    domains, Cross-entropy in general is a method of measuring the difference
    of information between two probability distributions (i.e. a way of
    measuring how similar they are). In our context, cross entropy is the
    difference between our observed values (which can be viewed as a
    probability distribution where every probability is either 0 or 1, since
    every value is known and thus has an extreme probability) and the
    predicted values (which are actual probabilities since the true values
    are unknown). Since we are interested in predicting values that follow a
    bernoulli distribution, the cross entropy takes the form of the negative
    log-likelihood of the bernoulli distribution.
    
    With this cost function defined, Neural Networks are just performing
    a more BA version of maximum likelihood estimation. A negative cost is
    defined because maximizing the likelihood is the same as minimizing the
    negative likelihood. 
    
    Parameters
    ----------
    y_true : numpy array
        True, observed values. The outcome of an event (where 1 == success).
    y_pred : numpy array
        The predicted success probabilities.
    delta : float, optional
        A small, positive constant to offset predicted probabilities that are
        zero, which avoids log(0). Is ignored if delta = 0.
        The default is 1e-9.

    Returns
    -------
    cost : float
        The binary cross-entropy.

    '''

    # Compute the cross-entropy cost
    # To avoid log(0) errors (not necessary in most cases)
    ypred = y_pred.copy()
    if delta != 0:
        ypred[ypred <= delta] = delta
        ypred[ypred >= 1-delta] = 1-delta
    
    # m is the number of observations, and m_scale is a scaling factor to make
    # the computation easier in case the gradients are really big. 
    
    cost = -np.sum(xlogy(y_true, ypred))
    return cost


def cross_entropy_binary_der(y_true, y_pred, delta=1e-9):
    '''
    The derivative of binary cross-entropy.
    
    The derivative of the cross-entropy function (i.e. derivative of the
    error output) with respect to the predicted value (y_pred).
    The value inside the brackets is the value of the derivative of the
    log-likelihood for a bernoulli distribution. The negative is added
    because of our definition of cross-entropy.

    Parameters
    ----------
    y_true : numpy array
        True, observed values. The outcome of an event (where 1 == success).
    y_pred : numpy array
        The predicted success probabilities.
    delta : float, optional
        A small, positive constant to offset predicted probabilities that are
        zero, which avoids log(0). Is ignored if delta = 0.
        The default is 1e-9.

    Returns
    -------
    float
        The derivative of binary cross-entropy.

    '''
    # Compute the cross-entropy cost
    # To avoid log(0) errors (not necessary in most cases)
    ypred = y_pred.copy()
    if delta != 0:
        ypred[ypred <= delta] = delta
        ypred[ypred >= 1-delta] = 1-delta
    
    return -((y_true/ypred) - (1-y_true)/(1-ypred))


def cross_entropy_der(y_true, y_pred, delta=1e-9):
    '''
    The derivative of binary cross-entropy.
    
    The derivative of the cross-entropy function (i.e. derivative of the
    error output) with respect to the predicted value (y_pred).
    The value inside the brackets is the value of the derivative of the
    log-likelihood for a bernoulli distribution. The negative is added
    because of our definition of cross-entropy.

    Parameters
    ----------
    y_true : numpy array
        True, observed values. The outcome of an event (where 1 == success).
    y_pred : numpy array
        The predicted success probabilities.
    delta : float, optional
        A small, positive constant to offset predicted probabilities that are
        zero, which avoids log(0). Is ignored if delta = 0.
        The default is 1e-9.

    Returns
    -------
    float
        The derivative of binary cross-entropy.

    '''
    # Compute the cross-entropy cost
    # To avoid log(0) errors (not necessary in most cases)
    ypred = y_pred.copy()
    if delta != 0:
        ypred[ypred <= delta] = delta
        ypred[ypred >= 1-delta] = 1-delta
    
    return -(y_true/ypred)


def softmax_cross_entropy_der(y_true, y_pred, z=None):
    return y_pred - y_true


def softmax_cross_entropy_der_fullmath(y_true, y_pred, z,
                                       delta=1e-9, center=True):
    dce = cross_entropy_der(y_true=y_true, y_pred=y_pred,
                            delta=delta)
    dout = softmax_der(x=z, center=center)
    dZ = dce.reshape(1, -1).dot(dout).reshape(y_true.shape[0], -1)
    return dZ


def sigmoid_cross_entropy_binary_der(y_true, y_pred, z,
                                       delta=1e-9, center=True):
    dce = cross_entropy_binary_der(y_true=y_true, y_pred=y_pred,
                                   delta=delta)
    dout = sigmoid_der(x=z)
    dZ = dce*dout
    return dZ


def mse(y_true, y_score):
    mse = np.sum((y_true - y_score)**2)
    return mse


def mse_der(y_true, y_score):
    der = -2*(y_true - y_score)
    return der


def linear(x):
    return x


def linear_der(x):
    return 1


def mse_linear_der(y_true, y_score, z):
    mseder = mse_der(y_true, y_score)
    # There was no activation function for the linear layer,
    # there is no derivative to compute here
    # linear_der = 1
    # mseder = mseder*linear_der
    return mseder



import numpy as np
import pandas as pd
from sklearn.utils import shuffle
from timeit import default_timer as dt

from .utils import print_time, updateProgBar


def downSample(data, target_var, majority=0, minority=1, bal=0.5,
               random_state=123):
    '''
    In a low base-rate classification problem, it is sometimes advantageous to downsample
    the majority class until the classes are balanced (or close enough). This 
    is done by randomly removing a proportion of observations from only the
    majority class.

    Parameters
    ----------
    data : pandas dataframe
        The design or feature matrix (with response).
    target_var : str
        The target or reponse variable.
    majority : int, optional
        Value of the majority class. The default is 0.
    minority : int, optional
        Value of the minority class. The default is 1.
    bal : float, optional
        The class balance after downsampling. The default is 0.5.
    random_state : int, optional
        The random seed for the process. The default is 123.

    Returns
    -------
    data_sub : pandas dataframe
        The downsampled data.

    '''
    data_min = data[data[target_var] == minority]
    data_maj = data[data[target_var] == majority]

    samp_size_train = int(len(data_min)/bal - len(data_min))
    
    np.random.seed(random_state)
    idx = np.random.choice(np.arange(0, len(data_maj)), samp_size_train,
                           replace=False)

    data_maj = data_maj.iloc[idx, :]
    
    data_sub = shuffle(pd.concat((data_min, data_maj), axis=0),
                       random_state=random_state)
    
    return data_sub


def genRandSampFromDF(data, sampSize, replace=False, random_state=None):
    '''
    Creates a Random sample of observations (rows) from a Dataframe (data). If replace=True,
    this function can be used for bootstrap samples. 

    Parameters
    ----------
    data : pandas dataframe
        The design or feature matrix (with response).
    sampSize : int
        Number of somples to return.
    replace : bool, optional
        Sample with replacement. The default is False.
    random_state : int, optional
        The random seed for the process. The default is None.

    Returns
    -------
    rand : pandas dataframe
        The randomly sampled data.

    '''
    if random_state is not None:
        np.random.seed(random_state)
    idx = list(np.random.choice(np.arange(0, data.shape[0]), sampSize, replace=replace))
    if isinstance(data, (pd.DataFrame, pd.Series)):
        rand = data.iloc[idx]
    else:
        rand = data[idx]
    return rand


def latinHypercube1D(data, sampleSize, random_state=None, shuffle_after=True,
                     sort_=True, sort_method="quicksort", sort_cols=None,
                     stratified=True, bin_placement="random", verbose=False):
    '''
    Creates a sample from a Dataframe (data). If replace=True,
    this function can be used for bootstrap samples. 

    Parameters
    ----------
    data : numpy array or pandas dataframe
        The design or feature matrix (with response).
    sampleSize : int
        The number of samples to return.
    random_state : int, optional
        The random seed of the process. The default is None.
    shuffle_after : bool, optional
        Shuffle the results after being sampled. The default is True.
    sort_ : bool, optional
        Sort the data (only set to False if sending presorted data).
        The default is True.
    sort_method : str, optional
        Numpy sort method. The default is "quicksort".
    sort_cols : list, optional
        The columns to include in the sorting. The default is None.
    stratified : bool, optional
        Create a stratified sample. The default is True.
    bin_placement : str, optional
        Method for placing the edges on the sampling bins.
        The default is "random".
    verbose : bool, optional
        Print the results of the process. The default is False.

    Raises
    ------
    ValueError
        Raised if invalid bin_plcement is passed.

    Returns
    -------
    LHC : pandas dataframe or numpy array
        The samples.

    '''
    t0 = None
    if verbose:
        t0 = dt()
        print_time("\nInitializing...", t0, te=dt())
    sortedData = None
    df = False
    if isinstance(data, pd.DataFrame):
        df = True
   
    if sort_:
        if verbose:
            print_time("\nSorting...", t0, te=dt())

        if df:
            sortedData = data.copy()
        else:
            sortedData = pd.DataFrame(data)

        if sort_cols is not None:
            sortedData = sortedData.sort_values(sort_cols, axis=0, kind=sort_method)
        else:
            sortedData = sortedData.sort_values(list(sortedData.columns), axis=0, kind=sort_method)
    
    if random_state is not None:
        np.random.seed(random_state)
        
    if verbose:
        print_time("\nShaping...", t0, te=dt())

    if sortedData is not None:
        sortedData = np.array(sortedData).reshape(data.shape[0], -1)
    else:
        sortedData = np.array(data).reshape(data.shape[0], -1)
    rows = sortedData.shape[0]
    cols = sortedData.shape[1]
    LHC = np.zeros(shape=(sampleSize, cols), dtype=data.dtypes)
    splits = sampleSize
    
    if verbose:
        print_time("\nCreating the bins...", t0, te=dt())
    
    if stratified:
        high = int(np.ceil(rows/sampleSize))
        low = int(np.floor(rows/sampleSize))
        rem = rows % sampleSize
        f_array = np.zeros(sampleSize)
        if rem != 0:
            rem2 = sampleSize - rem
            if bin_placement == "random":
                f_array = np.repeat((high, low), (rem, rem2))
                np.random.shuffle(f_array)
            elif bin_placement == "spaced":
                if rem > rem2:
                    r1 = np.repeat(high, rem)
                    a1 = np.arange(0, rem, np.floor(rem / rem2),
                                   dtype=np.int)[:rem2]
                    f_array = np.insert(r1, a1, low)
                else:
                    r1 = np.repeat(low, rem2)
                    a1 = np.arange(0, rem2, np.floor(rem2 / rem),
                                   dtype=np.int)[:rem]
                    f_array = np.insert(r1, a1, high)
            else:
                raise ValueError("""Not a valid bin placement. Change to 'spaced' or
                                 'random'. To order the bins from high to low, change
                                 stratified to False""")
        else:
            f_array = np.repeat(high, sampleSize)
        splits = np.cumsum(f_array)[:-1]
    Splits = np.array_split(sortedData, splits)
    nSplits = len(Splits)
    
    if verbose:
        print_time("\nSampling...", t0, te=dt())
       
    t1 = dt()
    for i, sample in enumerate(Splits):
        LHC[i, :] = sample[np.random.choice(sample.shape[0], 1)]
        if verbose:
            updateProgBar(i + 1, nSplits, t1)
    
    if shuffle_after:
        if verbose:
            print_time("\nShuffling...", t0, te=dt())
        np.random.shuffle(LHC)
    
    if df:
        if verbose:
            print_time("\nConverting to DataFrame...", t0, te=dt())
        LHC = pd.DataFrame(LHC, columns=data.columns).astype(data.dtypes)
        
    if verbose:
        print_time("\nFinished...", t0, te=dt())
    return LHC



from __future__ import (division, generators, absolute_import,
                        print_function, with_statement, nested_scopes,
                        unicode_literals)
import numpy as np

from scipy.optimize import minimize
from scipy.special import xlogy
from sklearn.base import BaseEstimator
from sklearn.linear_model import LogisticRegression as glm_lr
### keeping in case using an older version of sklearn
from statsmodels.genmod.generalized_linear_model import GLM as glm_pois
from statsmodels.genmod.families.family import Poisson
###
from sklearn.linear_model._base import LinearClassifierMixin
from sklearn.metrics import get_scorer
from sklearn.utils import check_array
from sklearn.utils.extmath import softmax
from sklearn.utils.validation import check_is_fitted, _check_sample_weight
from statsmodels.tools import add_constant
from timeit import default_timer as dt

from .distributions import (ap_logreg_lasso,
                                 posterior_logreg_lasso,
                                 likelihood_bernoulli,
                                 posterior_poisson_lasso,
                                 ap_poisson_lasso,
                                 posterior_poisson_lasso_od,
                                 ap_poisson_lasso_od,
                                 weibull_regression_post)
from .mcmc import adaptive_mcmc, rwm_with_lap, rwm, applyMCMC
from .tmath import mode_histogram, mode_kde
from .utils import print_time


class BayesLogRegClassifier(BaseEstimator, LinearClassifierMixin):
    '''
    A Logistic Regression Classifier that uses MCMC to evaluate the parameters.
    This objects inherits from sklearn\'s BaseEstimator and
    LinearClassifierMixin.
    '''
    def __init__(self, C=None, start=None, niter=10000, algo='rosenthal',
                 algo_options=None, retry_sd=0.02, retry_max_tries=100,
                 initialize_weights='sklearn', param_summary='mean',
                 has_constant=False, verbose=True,
                 over_dispersion=False, scorer=None):
        # TODO: Implement random_state for reporduceability
        '''
        The constructor for the BayesLogRegClassifier

        Parameters
        ----------
        C : float, optional
            The value for the inverse L1 penalty, or, the inverse
            regularization strength. If None, the MCMC process
            looks for the optimal penalty. The default is None.
            
            This value is gets converted into the scale parameter for a laplace
            distribution (scale = 2*C).
        start : numpy array (float), optional
            The starting values for the MCMC. If None, the MLE estimate is used
            (solved with sklearn). The default is None.
        niter : int, optional
            The number of MCMC samples to draw. The default is 10000.
        algo : str, optional
            The MCMC (Metropolis) algorithm to use. The default is 'rosenthal',
            which is a method that tunes the covariance matrix after each
            iteration. Other options include 'rwm', which is a simple random
            metropolis walk with a fixed covariance matrix, and 'lap' which is
            another adaptive method that tunes the covariance matrix every K
            iterations.
        algo_options : dict, optional
            The options to be passed to the MCMC algorithm. Include as a
            dictionary. The default is None.
        retry_sd : float, optional
            The MCMC alorithms use a Cholesky decomposition on the covariance
            matrix. In case the decomposition fails, the algorithms will 
            attempt to jitter the covaraince matrix to help it be positive 
            definite. This value determines the strength of the jittering and
            is drawn directly from a normal distribution with zero mean and
            retry_sd standard deviation. The default is 0.02.
        retry_max_tries : int, optional
            Number of attempts to correct the cholesky decomposition if it
            fails. The default is 100.
        initialize_weights : str, optional
            Determines the method of initializing the starting model parameter
            values (if start is None). Options are 'sklearn', 'ones', 'random',
            or 'zeros'. The default is 'sklearn'.
        param_summary : str, optional
            The method used in making the final parameter summaries. Options 
            are 'mean', 'median', 'mode_kde', 'mode_histogram', or
            'final_sample'. The default is 'mean'.
        has_constant : bool, optional
            Whether or not the data provided already has a column of ones
            as the first column in the dataset for the intercept of the model.
            If not, one is created. The default is False.
        verbose : bool, optional
            If True, a progress bar, along with timestamps, is provided.
            The default is True.
        over_dispersion : bool, optional
            Whether or not to account for overdispersion in the model.
            The default is False.
        scorer : function or str
            The function or string to use for the default scoring method.
            Otherwise, pass None for accuracy. The default is None.

        Returns
        -------
        None.

        '''
        self.C = C
        self.start = start
        self.niter = niter
        if algo == 'rosenthal':
            self.algo = adaptive_mcmc
        elif algo == 'lap':
            self.algo = rwm_with_lap
        else:
            self.algo = rwm
        self.algo_options = algo_options
        self.param_summary = param_summary
        self.initialize_weights = initialize_weights
        self.has_constant = has_constant
        self.verbose = verbose
        self.retry_sd = retry_sd
        self.retry_max_tries = retry_max_tries
        self.scorer = scorer
        self.over_dispersion = over_dispersion
        if C is None:
            # TODO: Implement over-dispersion
            #if over_dispersion:
            #    self.lpost = ap_logreg_lasso_od
            #    self.extra_params = 2
            #else:
            self.lpost = ap_logreg_lasso
            self.extra_params = 1
        else:
            #if over_dispersion:
            #    self.lpost = posterior_logreg_lasso_od
            #    self.extra_params = 1
            #else:
            self.lpost = posterior_logreg_lasso
            self.extra_params = 0
            
    def _create_coefs(self, mcmc_params, param_summary, extra_params):
        '''
        Creates Coefficients for MCMC model by aggregating the MCMC samples,
        using mean, median, mode, or last sampled value.

        Parameters
        ----------
        mcmc_params : numpy array (float)
            MCMC chain, one column for each parameter.
        param_summary : str
            The method of aggregation (either mean, median, mode_kde,
            mode_histogram, or last_value.
        extra_params : int
            The number of extra params used in the MCMC process (for example,
        parameters for priors or overdispersion).

        Returns
        -------
        coefs : numpy array (float)
            The coefficients of the model (aggregated from the MCMC chain).
        intercept : float
            The intercept of the model (aggregated from the MCMC chain).
        extra_parm : numpy array (float)
            Any extra parameters that were solved with the model (aggregated
            from the MCMC chain).

        '''
        sum_parms = None
        if param_summary == 'mean':
            sum_parms = mcmc_params.mean(axis=0)
        elif param_summary == 'median':
            sum_parms = np.median(mcmc_params, axis=0)
        elif param_summary == 'mode_histogram':
            sum_parms = np.array([mode_histogram(mcmc_params[:, i]) for i in range(mcmc_params.shape[1])])
        elif param_summary == 'mode_kde':
            sum_parms = np.array([mode_kde(mcmc_params[:, i]) for i in range(mcmc_params.shape[1])])
        else:
            sum_parms = mcmc_params[-1, :]
        
        if extra_params > 0:
            extra_parm = sum_parms[-extra_params:]
            coefs = sum_parms[1:-extra_params].reshape(1, -1)
        else:
            extra_parm = None
            coefs = sum_parms[1:].reshape(1, -1)
        intercept = np.array(sum_parms[0])
        
        return coefs, intercept, extra_parm
        
    def adjust_params_(self, new_param_summary):
        '''
        This method is intended to update the aggregated paramters with a new 
        summary as defined in new_param_summary. For example, if someone wanted
        to switch from coef_ representing the mean to coef_ representing 
        the median, they could use this function to do so.

        Parameters
        ----------
        new_param_summary : str
            The method of aggregating the MCMC chain parameters (either mean,
            median, mode_kde, mode_histogram, or last_value).

        Returns
        -------
        self
            Updates the coef_, intercept_, and extra_params_sum_ attributes.

        '''
        self.coef_, self.intercept_, self.extra_params_sum_ = self._create_coefs(self.mcmc_params, new_param_summary,
                                                                                 self.extra_params)
        return self
    
    def fit(self, X, y):
        '''
        Fit the model

        Parameters
        ----------
        X : numpy array
            The feature (or design) matrix.
        y : numpy array
            The response variable.

        Returns
        -------
        self
            Updates internal attributes, such as coef_ and intercept_.

        '''
        t0 = dt()
        X = check_array(X, force_all_finite='allow-nan', estimator=self, copy=True)
        if not self.has_constant:
            X = add_constant(X, prepend=True)       
        
        if self.start is not None:
            pass
        else:
            if self.verbose:
                print_time("Initializing Coefficients...", t0, dt(), backsn=True)
            if self.initialize_weights == 'sklearn':
                C = self.C
                if C is None:
                    C = 1
                mod = glm_lr(C=C, solver='liblinear', penalty='l1', fit_intercept=False).fit(X, y)
                self.start = mod.coef_.reshape(-1, )
                if self.extra_params > 0:
                    self.start = np.concatenate((self.start, np.repeat(1, self.extra_params)))
            elif self.initialize_weights == 'ones':
                self.start = np.ones(shape=X.shape[1] + self.extra_params)
            elif self.initialize_weights == 'random':
                self.start = np.random.normal(size=(X.shape[1] + self.extra_params, ))
            else:
                self.start = np.zeros(shape=X.shape[1] + self.extra_params)

        if self.verbose:
            print_time("Beginning MCMC...", t0, dt(), backsn=True)
        
        postArgs = {
            'X': X,
            'Y': y,
            'l_scale': self.C*2
        }
        
        algo_res = applyMCMC(st=self.start, ni=self.niter, lp=self.lpost,
                             algo=self.algo, postArgs=postArgs,
                             algoOpts=self.algo_options, sd=self.retry_sd,
                             max_tries=self.retry_max_tries)
        
        self.mcmc_params = algo_res['parameters']
        self.prev_vals = algo_res['prev_vals']
        
        self.coef_, self.intercept_, self.extra_params_sum_ = self._create_coefs(self.mcmc_params, self.param_summary,
                                                                                 self.extra_params)
        self.n_iter_ = self.niter
        self.classes_ = np.unique(y)
        
        if self.over_dispersion:
            self.dispersion_estimation_ = self.extra_params_sum_[-1]
        else:
            self.dispersion_estimation_ = None
        
        if self.verbose:
            print_time("Finished MCMC. Stored Coefficients...", t0, dt(), backsn=True)
            
        return self
    
    def predict(self, X):
        check_is_fitted(self, 'coef_')
        scores = self.decision_function(X)
        # note: decision_function returns X.dot(Betas). This code, modified from sklearn,
        # computes the decision based on the sign of the log odds ratio (X.dot(betas) == log(p/(1-p))).
        # When p <= 0.5, decision is 0, when p > 0.5, decision is 1 (threshold of 0.5 is the default for sklearn). 
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(np.int)
        else:
            indices = scores.argmax(axis=1)
        # returning the value in this way allows for someone to define their own classes,
        # for example, -1, 1 instead of 0, 1
        # Note: this is used in sklearn, but this code is not be robust enough for that,
        # so you need to define your classes as 0, 1
        return self.classes_[indices]
    
    def predict_proba(self, X):
        '''
        Generate predicted probability.

        Parameters
        ----------
        X : numpy array
            Feature (or design) matrix.

        Returns
        -------
        numpy array (float)
            The predicted probabilities.

        '''
        check_is_fitted(self, 'coef_')
        
        decision = self.decision_function(X)
        decision_2d = np.c_[-decision, decision]
        
        return softmax(decision_2d, copy=False)
    
    def predict_log_proba(self, X):
        '''
        Generate predicted log probability.

        Parameters
        ----------
        X : numpy array
            Feature (or design) matrix.

        Returns
        -------
        numpy array (float)
            The predicted log probabilities.

        '''
        return np.log(self.predict_proba(X))
    
    def score(self, X, y, sample_weight=None):
        '''
        Scores the model using the scoring method passed, or, the default
        scorer. In this case, the default scorer is accuracy.

        Parameters
        ----------
        X : numpy array or pandas dataframe
            The design or feature matrix.
        y : numpy array or pandas series
            The target or response variable.
        sample_weight : numpy array, optional
            An array containing the weights for each sample.
            The default is None.

        Returns
        -------
        score
            The result of the scoring function.

        '''
        scorer = self.scorer or 'accuracy'
        scorer = get_scorer(scorer)
        
        return scorer(self, X, y, sample_weight=sample_weight)


class BayesPoissonRegressor(BaseEstimator, LinearClassifierMixin):
    '''
    A Poisson Regressor that uses MCMC to evaluate the parameters.
    This objects inherits from sklearn\'s BaseEstimator and
    LinearClassifierMixin.
    '''
    def __init__(self, C=1, start=None, niter=10000, algo='rosenthal',
                 algo_options=None, retry_sd=0.02, retry_max_tries=100,
                 initialize_weights='sklearn', param_summary='mean',
                 has_constant=False, verbose=True,
                 over_dispersion=False, scorer='D2'):
        '''
        The constructor for the BayesPoissonClassifier

        Parameters
        ----------
        C : float, optional
            The value for the inverse L1 penalty, or, the inverse
            regularization strength. If None, the MCMC process
            looks for the optimal penalty. The default is None.
            
            This value is gets converted into the scale parameter for a laplace
            distribution (scale = 2*C).
        start : numpy array (float), optional
            The starting values for the MCMC. If None, the MLE estimate is used
            (solved with sklearn). The default is None.
        niter : int, optional
            The number of MCMC samples to draw. The default is 10000.
        algo : str, optional
            The MCMC (Metropolis) algorithm to use. The default is 'rosenthal', which is 
            a method that tunes the covariance matrix after each iteration.
            Other options include 'rwm', which is a simple random metropolis
            walk with a fixed covariance matrix, and 'lap' which is another 
            adaptive method that tunes the covariance matrix every K
            iterations.
        algo_options : dict, optional
            The options to be passed to the MCMC algorithm. Include as a
            dictionary. The default is None.
        retry_sd : float, optional
            The MCMC alorithms use a Cholesky decomposition on the covariance
            matrix. In case the decomposition fails, the algorithms will 
            attempt to jitter the covaraince matrix to help it be positive 
            definite. This value determines the strength of the jittering and
            is drawn directly from a normal distribution with zero mean and
            retry_sd standard deviation. The default is 0.02.
        retry_max_tries : int, optional
            Number of attempts to correct the cholesky decomposition if it
            fails. The default is 100.
        initialize_weights : str, optional
            Determines the method of initializing the starting model parameter
            values (if start is None). Options are 'sklearn', 'ones', 'random',
            or 'zeros'. The default is 'sklearn'.
        param_summary : str, optional
            The method used in making the final parameter summaries. Options 
            are 'mean', 'median', 'mode_kde', 'mode_histogram', or
            'final_sample'. The default is 'mean'.
        has_constant : bool, optional
            Whether or not the data provided already has a column of ones
            as the first column in the dataset for the intercept of the model.
            If not, one is created. The default is False.
        verbose : bool, optional
            If True, a progress bar, along with timestamps, is provided.
            The default is True.
        over_dispersion : bool, optional
            Whether or not to account for overdispersion in the model.
            The default is False.

        Returns
        -------
        None.

        '''
        self.C = C
        self.start = start
        self.niter = niter
        if algo == 'rosenthal':
            self.algo = adaptive_mcmc
        elif algo == 'lap':
            self.algo = rwm_with_lap
        else:
            self.algo = rwm
        self.algo_options = algo_options
        self.param_summary = param_summary
        self.initialize_weights = initialize_weights
        self.has_constant = has_constant
        self.verbose = verbose
        self.retry_sd = retry_sd
        self.retry_max_tries = retry_max_tries
        self.over_dispersion = over_dispersion
        self.scorer = scorer
        if C is None:
            if over_dispersion:
                self.lpost = ap_poisson_lasso_od
                self.extra_params = 2
            else:
                self.lpost = ap_poisson_lasso
                self.extra_params = 1
        else:
            if over_dispersion:
                print("using OD")
                self.lpost = posterior_poisson_lasso_od
                self.extra_params = 1
            else:
                self.lpost = posterior_poisson_lasso
                self.extra_params = 0
    
    def _deviance_dispersion_update(self, X, y, sample_weight=None):
        weights = _check_sample_weight(sample_weight, X)
        y_pred = self.predict(X)
        y_mean = np.average(y, weights=weights)
        deviance_ = np.sum(weights * (2*(xlogy(y, y/y_pred) - y + y_pred)))
        null_deviance_ = np.sum(weights * (2*(xlogy(y, y/y_mean) - y + y_mean)))
        # pearson residual:  (raw residual)/(variance function)
        pearson_residuals_ = (y - y_pred)/np.sqrt(y_pred)
        pearson_chi2_ = np.sum(pearson_residuals_**2)
        model_d2_ = 1 - deviance_/null_deviance_
        # degrees of freedom of the model (all params (including intercept) minus 1)
        df_model_ = X.shape[1]
        # degrees of freedom of residuals ((n_obs - 1) - (nparms - 1)), or
        df_residuals_ = X.shape[0] - X.shape[1] - 1
        # total degrees of freedom
        df_total_ = df_residuals_ + df_model_
        # method of moments estimator for dispersion scale
        dispersion_scale_ = pearson_chi2_/df_residuals_
        dispersion_scale_sqrt_ = np.sqrt(dispersion_scale_)
        results = {'deviance_': deviance_,
                   'null_deviance_': null_deviance_,
                   'pearson_residuals_': pearson_residuals_,
                   'pearson_chi2_': pearson_chi2_,
                   'model_d2_': model_d2_,
                   'df_model_': df_model_,
                   'df_residuals_': df_residuals_,
                   'df_total_': df_total_,
                   'dispersion_scale_': dispersion_scale_,
                   'dispersion_scale_sqrt_': dispersion_scale_sqrt_}
        return results
        
    def _create_coefs(self, mcmc_params, param_summary, extra_params):
        '''
        Creates Coefficients for MCMC model by aggregating the MCMC samples,
        using mean, median, mode, or last sampled value.

        Parameters
        ----------
        mcmc_params : numpy array (float)
            MCMC chain, one column for each parameter.
        param_summary : str
            The method of aggregation (either mean, median, mode_kde,
            mode_histogram, or last_value.
        extra_params : int
            The number of extra params used in the MCMC process (for example,
        parameters for priors or overdispersion).

        Returns
        -------
        coefs : numpy array (float)
            The coefficients of the model (aggregated from the MCMC chain).
        intercept : float
            The intercept of the model (aggregated from the MCMC chain).
        extra_parm : numpy array (float)
            Any extra parameters that were solved with the model (aggregated
            from the MCMC chain).

        '''
        sum_parms = None
        if param_summary == 'mean':
            sum_parms = mcmc_params.mean(axis=0)
        elif param_summary == 'median':
            sum_parms = np.median(mcmc_params, axis=0)
        elif param_summary == 'mode_histogram':
            sum_parms = np.array([mode_histogram(mcmc_params[:, i]) for i in range(mcmc_params.shape[1])])
        elif param_summary == 'mode_kde':
            sum_parms = np.array([mode_kde(mcmc_params[:, i]) for i in range(mcmc_params.shape[1])])
        else:
            sum_parms = mcmc_params[-1, :]
        
        if extra_params > 0:
            extra_parm = sum_parms[-extra_params:]
            coefs = sum_parms[1:-extra_params].reshape(-1, )
        else:
            extra_parm = None
            coefs = sum_parms[1:].reshape(-1, )
        intercept = np.array(sum_parms[0])
        
        return coefs, intercept, extra_parm
        
    def adjust_params_(self, new_param_summary):
        '''
        This method is intended to update the aggregated paramters with a new 
        summary as defined in new_param_summary. For example, if someone wanted
        to switch from coef_ representing the mean to coef_ representing 
        the median, they could use this function to do so.

        Parameters
        ----------
        new_param_summary : str
            The method of aggregating the MCMC chain parameters (either mean,
            median, mode_kde, mode_histogram, or last_value).

        Returns
        -------
        self
            Updates the coef_, intercept_, and extra_params_sum_ attributes.

        '''
        self.coef_, self.intercept_, self.extra_params_sum_ = self._create_coefs(self.mcmc_params, new_param_summary,
                                                                                 self.extra_params)
        
        return self
    
    def fit(self, X, y):
        '''
        Fit the model

        Parameters
        ----------
        X : numpy array
            The feature (or design) matrix.
        y : numpy array
            The response variable.

        Returns
        -------
        self
            Updates internal attributes, such as coef_ and intercept_.

        '''
        t0 = dt()
        X = check_array(X, force_all_finite='allow-nan', estimator=self, copy=True)
        if not self.has_constant:
            X = add_constant(X, prepend=True)       
        
        if self.start is not None:
            pass
        else:
            if self.verbose:
                print_time("Initializing Coefficients...", t0, dt(), backsn=True)
            if self.initialize_weights == 'sklearn':
                C = self.C
                if C is None:
                    C = 1
                try:
                    from sklearn.linear_model import PoissonRegressor as glm_pois_sk
                    mod = glm_pois_sk(alpha=1/C, fit_intercept=False, max_iter=1000).fit(X, y)
                    self.start = mod.coef_.reshape(-1, )
                except ImportError:
                    print('Older sklearn, no PoissonRegressor. Using statsmodels instead')
                    mod = glm_pois(y, X, family=Poisson()).fit()
                    self.start = mod.params.reshape(-1, )
                if self.extra_params > 0:
                    self.start = np.concatenate((self.start, np.repeat(1, self.extra_params)))
            elif self.initialize_weights == 'ones':
                self.start = np.ones(shape=X.shape[1] + self.extra_params)
            elif self.initialize_weights == 'random':
                self.start = np.random.normal(size=(X.shape[1] + self.extra_params, ))
            else:
                self.start = np.zeros(shape=X.shape[1] + self.extra_params)

        if self.verbose:
            print_time("Beginning MCMC...", t0, dt(), backsn=True)
        
        postArgs = {
            'X': X,
            'Y': y,
            'l_scale': self.C*2
        }
        
        algo_res = applyMCMC(st=self.start, ni=self.niter, lp=self.lpost,
                             algo=self.algo, postArgs=postArgs,
                             algoOpts=self.algo_options, sd=self.retry_sd,
                             max_tries=self.retry_max_tries)
        
        self.mcmc_params = algo_res['parameters']
        self.prev_vals = algo_res['prev_vals']
        
        self.coef_, self.intercept_, self.extra_params_sum_ = self._create_coefs(self.mcmc_params, self.param_summary,
                                                                                 self.extra_params)
        self.n_iter_ = self.niter
        
        # get model summaries
        if self.over_dispersion:
            self.dispersion_delta_ = self.extra_params_sum_[-1]
            self.dispersion_estimation_ = 1/(1 - self.dispersion_delta_)**2
        else:
            self.dispersion_delta_ = 0
            self.dispersion_estimation_ = None
        
        ddu = self._deviance_dispersion_update(X[:, 1:], y,
                                               sample_weight=None)
        self.deviance_ = ddu['deviance_']
        self.null_deviance_ = ddu['null_deviance_']
        self.pearson_residuals_ = ddu['pearson_residuals_']
        self.pearson_chi2_ = ddu['pearson_chi2_']
        self.model_d2_ = ddu['model_d2_']
        self.df_model_ = ddu['df_model_']
        self.df_residuals_ = ddu['df_residuals_']
        self.df_total_ = ddu['df_total_']
        self.dispersion_scale_ = ddu['dispersion_scale_']
        self.dispersion_scale_sqrt_ = ddu['dispersion_scale_sqrt_']
        
        return self
    
    def predict(self, X):
        check_is_fitted(self, 'coef_')
        X = check_array(X, accept_sparse=['csr', 'csc', 'coo'],
                        dtype=[np.float64, np.float32], ensure_2d=True,
                        allow_nd=False)
        mu = X @ self.coef_ + self.intercept_
        #return np.exp(mu)
        return np.exp(mu)*(1-self.dispersion_delta_)
    
    def score(self, X, y, sample_weight=None):
        if self.scorer == 'D2':
            ddu = self._deviance_dispersion_update(X, y,
                                                   sample_weight=None)
            score = ddu['model_d2_']
        else:
            score = self.scorer(X, y, sample_weight=sample_weight)
        
        return score

    
class BayesWeibullRegressor(BaseEstimator, LinearClassifierMixin):
    '''
    A Weibull Regressor that uses MCMC to evaluate the parameters.
    This objects inherits from sklearn\'s BaseEstimator and
    LinearClassifierMixin.
    '''
    def __init__(self, C=1, start=None, niter=10000, algo='rosenthal',
                 algo_options=None, retry_sd=0.02, retry_max_tries=100,
                 initialize_weights='sklearn', param_summary='mean',
                 has_constant=False, verbose=True,
                 over_dispersion=False):
        '''
        The constructor for the BayesWeibullClassifier

        Parameters
        ----------
        C : float, optional
            The value for the inverse L1 penalty, or, the inverse
            regularization strength. If None, the MCMC process
            looks for the optimal penalty. The default is None.
            
            This value is gets converted into the scale parameter for a laplace
            distribution (scale = 2*C).
        start : numpy array (float), optional
            The starting values for the MCMC. If None, the MLE estimate is used
            (solved with sklearn). The default is None.
        niter : int, optional
            The number of MCMC samples to draw. The default is 10000.
        algo : str, optional
            The MCMC (Metropolis) algorithm to use. The default is 'rosenthal', which is 
            a method that tunes the covariance matrix after each iteration.
            Other options include 'rwm', which is a simple random metropolis
            walk with a fixed covariance matrix, and 'lap' which is another 
            adaptive method that tunes the covariance matrix every K
            iterations.
        algo_options : dict, optional
            The options to be passed to the MCMC algorithm. Include as a
            dictionary. The default is None.
        retry_sd : float, optional
            The MCMC alorithms use a Cholesky decomposition on the covariance
            matrix. In case the decomposition fails, the algorithms will 
            attempt to jitter the covaraince matrix to help it be positive 
            definite. This value determines the strength of the jittering and
            is drawn directly from a normal distribution with zero mean and
            retry_sd standard deviation. The default is 0.02.
        retry_max_tries : int, optional
            Number of attempts to correct the cholesky decomposition if it
            fails. The default is 100.
        initialize_weights : str, optional
            Determines the method of initializing the starting model parameter
            values (if start is None). Options are 'sklearn', 'ones', 'random',
            or 'zeros'. The default is 'sklearn'.
        param_summary : str, optional
            The method used in making the final parameter summaries. Options 
            are 'mean', 'median', 'mode_kde', 'mode_histogram', or
            'final_sample'. The default is 'mean'.
        has_constant : bool, optional
            Whether or not the data provided already has a column of ones
            as the first column in the dataset for the intercept of the model.
            If not, one is created. The default is False.
        verbose : bool, optional
            If True, a progress bar, along with timestamps, is provided.
            The default is True.
        over_dispersion : bool, optional
            ---CURRENTLY NOT IMPLEMENTED---
            Whether or not to account for overdispersion in the model.
            The default is False.

        Returns
        -------
        None.

        '''
        self.C = C
        self.start = start
        self.niter = niter
        if algo == 'rosenthal':
            self.algo = adaptive_mcmc
        elif algo == 'lap':
            self.algo = rwm_with_lap
        else:
            self.algo = rwm
        self.algo_options = algo_options
        self.param_summary = param_summary
        self.initialize_weights = initialize_weights
        self.has_constant = has_constant
        self.verbose = verbose
        self.retry_sd = retry_sd
        self.retry_max_tries = retry_max_tries
        self.lpost = weibull_regression_post
        self.extra_params = 0
            
    def _create_coefs(self, mcmc_params, param_summary, extra_params):
        '''
        Creates Coefficients for MCMC model by aggregating the MCMC samples,
        using mean, median, mode, or last sampled value.

        Parameters
        ----------
        mcmc_params : numpy array (float)
            MCMC chain, one column for each parameter.
        param_summary : str
            The method of aggregation (either mean, median, mode_kde,
            mode_histogram, or last_value.
        extra_params : int
            The number of extra params used in the MCMC process (for example,
        parameters for priors or overdispersion).

        Returns
        -------
        coefs : numpy array (float)
            The coefficients of the model (aggregated from the MCMC chain).
        intercept : float
            The intercept of the model (aggregated from the MCMC chain).
        extra_parm : numpy array (float)
            Any extra parameters that were solved with the model (aggregated
            from the MCMC chain).

        '''
        sum_parms = None
        if param_summary == 'mean':
            sum_parms = mcmc_params.mean(axis=0)
        elif param_summary == 'median':
            sum_parms = np.median(mcmc_params, axis=0)
        elif param_summary == 'mode_histogram':
            sum_parms = np.array([mode_histogram(mcmc_params[:, i]) for i in range(mcmc_params.shape[1])])
        elif param_summary == 'mode_kde':
            sum_parms = np.array([mode_kde(mcmc_params[:, i]) for i in range(mcmc_params.shape[1])])
        else:
            sum_parms = mcmc_params[-1, :]
        
        if extra_params > 0:
            extra_parm = sum_parms[-extra_params:]
            coefs = sum_parms[1:-extra_params].reshape(-1, )
        else:
            extra_parm = None
            coefs = sum_parms[1:].reshape(-1, )
        intercept = np.array(sum_parms[0])
        
        return coefs, intercept, extra_parm
    
    def _deviance_dispersion_update(self, X, y, sample_weight=None):
        weights = _check_sample_weight(sample_weight, X)
        y_pred = self.predict(X)
        y_mean = np.average(y, weights=weights)
        deviance_ = np.sum(weights * (2*(xlogy(y, y/y_pred) - y + y_pred)))
        null_deviance_ = np.sum(weights * (2*(xlogy(y, y/y_mean) - y + y_mean)))
        # pearson residual:  (raw residual)/(variance function)
        # TODO: put correct weibull variance here
        pearson_residuals_ = (y - y_pred)/np.sqrt(y_pred)
        pearson_chi2_ = np.sum(pearson_residuals_**2)
        model_d2_ = 1 - deviance_/null_deviance_
        # degrees of freedom of the model (all params (including intercept) minus 1)
        df_model_ = X.shape[1]
        # degrees of freedom of residuals ((n_obs - 1) - (nparms - 1)), or
        df_residuals_ = X.shape[0] - X.shape[1] - 1
        # total degrees of freedom
        df_total_ = df_residuals_ + df_model_
        # method of moments estimator for dispersion scale
        dispersion_scale_ = pearson_chi2_/df_residuals_
        dispersion_scale_sqrt_ = np.sqrt(dispersion_scale_)
        results = {'deviance_': deviance_,
                   'null_deviance_': null_deviance_,
                   'pearson_residuals_': pearson_residuals_,
                   'pearson_chi2_': pearson_chi2_,
                   'model_d2_': model_d2_,
                   'df_model_': df_model_,
                   'df_residuals_': df_residuals_,
                   'df_total_': df_total_,
                   'dispersion_scale_': dispersion_scale_,
                   'dispersion_scale_sqrt_': dispersion_scale_sqrt_}
        return results
        
    def adjust_params_(self, new_param_summary):
        '''
        This method is intended to update the aggregated paramters with a new 
        summary as defined in new_param_summary. For example, if someone wanted
        to switch from coef_ representing the mean to coef_ representing 
        the median, they could use this function to do so.

        Parameters
        ----------
        new_param_summary : str
            The method of aggregating the MCMC chain parameters (either mean,
            median, mode_kde, mode_histogram, or last_value).

        Returns
        -------
        self
            Updates the coef_, intercept_, and extra_params_sum_ attributes.

        '''
        self.coef_, self.intercept_, self.extra_params_sum_ = self._create_coefs(self.mcmc_params, new_param_summary,
                                                                                 self.extra_params)
        return self
    
    def fit(self, X, y):
        '''
        Fit the model

        Parameters
        ----------
        X : numpy array
            The feature (or design) matrix.
        y : numpy array
            The response variable.

        Returns
        -------
        self
            Updates internal attributes, such as coef_ and intercept_.

        '''
        t0 = dt()
        X = check_array(X, force_all_finite='allow-nan', estimator=self, copy=True)
        if not self.has_constant:
            X = add_constant(X, prepend=True)       
        
        if self.start is not None:
            pass
        else:
            if self.verbose:
                print_time("Initializing Coefficients...", t0, dt(), backsn=True)
            if self.initialize_weights == 'sklearn':
                C = self.C
                if C is None:
                    C = 1
                # TODO: update with Weibull Regression starting values
                ### If using sklearn version 0.23.2, can use this line instead
                #mod = glm_pois(alpha=1/C, fit_intercept=False, max_iter=1000).fit(X, y)
                #self.start = mod.coef_.reshape(-1, )
                ### else, use statsmodels
                mod = glm_pois(y, X, family=Poisson()).fit()
                self.start = mod.params.reshape(-1, )
                if self.extra_params > 0:
                    self.start = np.concatenate((self.start, np.repeat(1, self.extra_params)))
            elif self.initialize_weights == 'ones':
                self.start = np.ones(shape=X.shape[1] + self.extra_params)
            elif self.initialize_weights == 'random':
                self.start = np.random.normal(X.shape[1] + self.extra_params)
            else:
                self.start = np.zeros(shape=X.shape[1] + self.extra_params)

        if self.verbose:
            print_time("Beginning MCMC...", t0, dt(), backsn=True)
        
        postArgs = {
            'X': X,
            'Y': y,
            'l_scale': self.C*2
        }
        
        algo_res = applyMCMC(st=self.start, ni=self.niter, lp=self.lpost,
                             algo=self.algo, postArgs=postArgs,
                             algoOpts=self.algo_options, sd=self.retry_sd,
                             max_tries=self.retry_max_tries)
        
        self.mcmc_params = algo_res['parameters']
        self.prev_vals = algo_res['prev_vals']
        
        self.coef_, self.intercept_, self.extra_params_sum_ = self._create_coefs(self.mcmc_params, self.param_summary,
                                                                                 self.extra_params)
        self.n_iter_ = self.niter
        
        #get model summaries
        weights = _check_sample_weight(None, X)
        y_pred = self.predict(X[:, 1:])
        y_mean = np.average(y, weights=weights)
        dev = np.sum(weights * (2*(xlogy(y, y/y_pred) - y + y_pred)))
        dev_null = np.sum(weights * (2*(xlogy(y, y/y_mean) - y + y_mean)))
        self.deviance_ = dev
        self.null_deviance_ = dev_null
        self.pearson_residuals_ = (y - y_pred)/np.sqrt(y_pred)
        self.pearson_chi2_ = np.sum(self.pearson_residuals_**2)
        self.model_d2_ = 1 - dev/dev_null
        self.df_model_ = X.shape[1] - 1
        self.df_residuals_ = X.shape[0] - X.shape[1]
        self.dispersion_scale_ = self.pearson_chi2_/self.df_residuals_
        self.dispersion_scale_sqrt_ = np.sqrt(self.dispersion_scale_)
        
        return self
    
    def predict(self, X):
        check_is_fitted(self, 'coef_')
        X = check_array(X, accept_sparse=['csr', 'csc', 'coo'],
                        dtype=[np.float64, np.float32], ensure_2d=True,
                        allow_nd=False)
        mu = X @ self.coef_ + self.intercept_
        return np.exp(mu)
    
    def score(self, X, y, sample_weight=None, scorer='D2'):
        if scorer == 'D2':
            ddu = self._deviance_dispersion_update(X[:, 1:], y,
                                                   sample_weight=None)
            score = 1 - ddu['deviance_']/ddu['null_deviance_']
        else:
            score = scorer(X, y, sample_weight=sample_weight)
        
        return score
    

class LogReg(BaseEstimator, LinearClassifierMixin):
    '''
    A class for performing Logistic Regression using scipy.optimize.minimize.
    This is not meant to replace skleaarn's implementation, and in fact, it
    is mainly built on sklearn. This was mainly a test for the author to get a
    better understanding of both the internals of sklearn and
    Logistic Regression
    '''
    def __init__(self, x0=None, lamb=1, l_norm=1, method=None, jac=None,
                 hess=None, hessp=None, bounds=None, constraints=(), tol=None,
                 callback=None, options=None, has_constant=False):
        
        self.optimize_args = {'fun': likelihood_bernoulli,
                              'x0': x0,
                              'method': method,
                              'jac': jac,
                              'hess': hess,
                              'hessp': hessp,
                              'bounds': bounds,
                              'constraints': constraints,
                              'tol': tol,
                              'callback': callback,
                              'options': options
                             }
        self.lamb = lamb
        self.l_norm = l_norm
        
        
    def fit(self, X, y):
        '''
        Fit the model

        Parameters
        ----------
        X : numpy array
            The feature (or design) matrix.
        y : numpy array
            The response variable.

        Returns
        -------
        self
            Updates internal attributes, such as coef_ and intercept_.

        '''
        
        if not self.has_constant:
            X = add_constant(X, prepend=True)
        
        if self.optimize_args['x0'] is None:
            self.optimize_args['x0'] = np.zeros(shape=X.shape[1])
            #self.optimize_args['x0'] = np.random.normal(size=X.shape[1])
        
        args = (X, y, self.lamb, self.l_norm)
        
        self.optimize_args.update({'args': args})
        self.optim_results = minimize(**self.optimize_args)
        
        self.intercept_ = self.optim_results.x[0].reshape(1, )
        self.coef_ = self.optim_results.x[1:].reshape(1, -1)
        self.n_iter_ = self.optim_results.nit
        self.classes_ = np.unique(y)
        return self
    
    def predict(self, X):
        scores = self.decision_function(X)
        # note: decision_function returns X.dot(Betas). This code, modified from sklearn,
        # computes the decision based on the sign of the log odds ratio (X.dot(betas) == log(p/(1-p))).
        # When p <= 0.5, decision is 0, when p > 0.5, decision is 1 (threshold of 0.5 is the default for sklearn). 
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(np.int)
        else:
            indices = scores.argmax(axis=1)
        # returning the value in this way allows for someone to define their own classes,
        # for example, -1, 1 instead of 0, 1
        # Note: this is used in sklearn, but this code is not be robust enough for that,
        # so you need to define your classes as 0, 1
        return self.classes_[indices]
    
    def predict_proba(self, X):
        '''
        Generate predicted probability.

        Parameters
        ----------
        X : numpy array
            Feature (or design) matrix.

        Returns
        -------
        numpy array (float)
            The predicted probabilities.

        '''
        #check_is_fitted(self)
        
        decision = self.decision_function(X)
        decision_2d = np.c_[-decision, decision]
        
        return softmax(decision_2d, copy=False)
    
    def predict_log_proba(self, X):
        return np.log(self.predict_proba(X))
    
    def score(self, X, y, sample_weight=None):
        #TODO: Test default scoring
        scoring = self.scoring or 'accuracy'
        scoring = get_scorer(scoring)
        
        return scoring(self, X, y, sample_weight=sample_weight)